import spacepy
import spacepy.pybats

import plotly
import plotly.graph_objs as go
from plotly.offline import init_notebook_mode, iplot, plot
init_notebook_mode(connected = True)

import numpy as np
import pandas as pd

import scipy.interpolate as spint
import scipy.spatial.qhull as qhull

from kamodo import Kamodo, kamodofy, gridify,pointlike
from readers.OCTREE_BLOCK_GRID._interpolate_amrdata import ffi as OCTREE_BLOCK_GRID_FFI
from readers.OCTREE_BLOCK_GRID._interpolate_amrdata import lib as OCTREE_BLOCK_GRID_LIB

import time
import glob


import os
from os import path
from datetime import datetime, timedelta

class SWMF_GM(Kamodo): 
    def __init__(self, 
                 filename, 
                 runpath = "./", 
                 runname = "noname", 
                 start_time = "1000/01/01 00:00",
                 setup_interpolators=True,
                 variables_requested = None,
                 **kwargs):
        # Start timer
        tic = time.perf_counter()

        # Prepare model for function registration
        super(SWMF_GM, self).__init__(**kwargs) 

        # perform any necessary I/O
        print('opening {}'.format(filename))
        self.filename = filename
        self.runpath = runpath
        self.runname = runname
        self.start_time = start_time
        self.missing_value = np.NAN
        self.variables=dict()
        mhd = spacepy.pybats.IdlFile(filename,sort_unstructured_data=False) #- spacepy/pybats/__init__.py modified to not sort by default when reading binary *.out IDL files

        toc1 = time.perf_counter()

        # Pull out variables and units
        if variables_requested is None:
            vars = list(mhd.keys())
        else:
            vars=['grid','x','y','z']
            mhd_keys=list(mhd.keys())
            print('mhd_keys:',mhd_keys)
            for varname in variables_requested:
                print(varname)
                if varname in mhd_keys:
                    print('appending',varname)
                    vars.append(varname)
        print(vars)
        self.var_keys=mhd.keys()
        vars = vars[4:]
        self.nvars = len(vars)
        self.vars = vars
        units = mhd.meta['header'].split("_", 1)[0].strip().split(" ")
        units = units[3:]
        
        # var = vars[0]
        # unit = units[vars.index(var)]
        # self.nvars = mhd.meta['nvar']
        print('... found',self.nvars,'variables:',vars,units)
#        print(vars)
        self.var = vars[0]
        for var in vars:
            unit=units[vars.index(var)].replace("Mp/cc","1/cm**3").replace("m3","m**3").replace("m2","m**2").replace("uA","muA").replace("--","")
            
            self.variables[var] = dict(units = unit, data = np.array(mhd[var], dtype=np.float32) )

        
#        self.variables[var] = dict(units = units[vars.index(var)], data = np.array(mhd[var], dtype=np.float32) )
        
        # Pull start_time from the DatabaseInfo file (if it exists), then add simulated time from file metadata
        if path.isfile(runpath+runname+'/DatabaseInfo'):
            for line in open(runpath+runname+'/DatabaseInfo'):
                if "start_time" in line:
                    self.start_time = line.split("#", 1)[0].strip()
                    print('... found run start_time = ',self.start_time)
                    break

        dt = datetime.strptime(self.start_time, '%Y/%m/%d %H:%M')
        dt = dt + timedelta(seconds=int(mhd.meta['runtime']))
        self.filetime = dt.strftime("%Y/%m/%d %H:%M:%S UT")
        print('... setting filetime = ',self.filetime)

        toc2 = time.perf_counter()

        self.meta=mhd.meta
        
        # pull grid information into C program extracting Octree information
        # x=mhd['x']
        # y=mhd['y']
        # z=mhd['z']
        # 
        # Pull grid to interpolate from into stored arrays
        self.grid = np.empty([mhd['x'].size, 3], dtype=np.float32)
        self.grid[:,0] = np.array(mhd['x'])
        self.grid[:,1] = np.array(mhd['y'])
        self.grid[:,2] = np.array(mhd['z'])
        self.x=mhd['x']
        self.y=mhd['y']
        self.z=mhd['z']
        self.Ncell=len(self.x)
        dd = np.absolute(np.concatenate(( np.array(mhd['x']), np.array([999.], dtype=np.float32) )) - 
                         np.concatenate(( np.array([999.], dtype=np.float32), np.array(mhd['x']) )) )
        self.gridSize = mhd['x'].size
        self.gridMinDx = np.amin(dd[dd[:]>0.])
        print('... file grid has',self.gridSize,'cells with minimum dx =',self.gridMinDx)
        
        
        # Interpolation values
        self.plots = dict()
        self.plottype = "XY"
        self.cut = 'Z'
        self.cutV = 0.
        self.nX = 141
        self.nY = 81
        self.nZ = 1
        self.newx = np.linspace(-50., 20., self.nX)
        self.newy = np.linspace(-20., 20., self.nY)
        self.newz = self.cutV

        self.tol = 1.1
        self.plots[self.plottype] = dict(cut=self.cut, cutV=self.cutV, tol=self.tol,
                                         nX=self.nX, nY=self.nY, nZ=self.nZ,
                                         newx=self.newx, newy=self.newy, newz=self.newz)  
        toc3=None
        toc4=None
        toc3 = time.perf_counter()
        # Get grids ready to use
        #            self.setup_interpolating_grids()
        self.setup_octree()
        toc4 = time.perf_counter()
        print(self.variables)
        for varname in self.variables:
            units = self.variables[varname]['units']
            self.register_variable(varname, units)
            self.plots[self.plottype][varname]=self.variables[varname]['interpolator']
            print(varname,units)
                
        if 'bx' in self.variables and 'by' in self.variables and 'bz' in self.variables:
            def Bvec(xvec):
                bx=self.bx(xvec)
                by=self.by(xvec)
                bz=self.bz(xvec)
                return((np.vstack([bx,by,bz])).T)

            print("adding Bvec ...")
            interpolator = Bvec
            units=self.variables['bx']['units']
            self['Bvec'] = kamodofy(interpolator, 
                                 units = units, 
                                 citation = "Rastaetter 2021",
                                 data = None)

            @kamodofy(equation = "\\hat{n}(\\vec{y}) = \\vec{y}/\\sqrt{\\vec{y} \\cdot \\vec{y}}")
            @pointlike(signature = '(m,n)->(m,n)', squeeze = 0)
            def normalized(yvec):   
                r = np.linalg.norm(yvec, axis = 1)
                r[r==0] = np.nan
                
                try:
                    return yvec/r
                except:
                    return (yvec.T/r).T


            self['nhat'] = normalized

            self['bhat']= "nhat(Bvec)"
            
        else:
            print("NOT adding Bvec")

        self.set_plot(plottype='XZ')

        # end timer
        toc = time.perf_counter()
        print(f"Time loading file: {toc1 - tic:0.4f} seconds")
        print(f"Time loading file defining dictionaries for Kamodo: {toc2 - tic:0.4f} seconds")
        if toc3 is not None:
            print(f"Time loading file and kamodofications: {toc3 - tic:0.4f} seconds")
        if toc4 is not None:
            print(f"Time loading file setup interpolating grids: {toc4 - tic:0.4f} seconds")
        print(f"Time loading file full kamodofication and and precomputing default interpolations: {toc - tic:0.4f} seconds")
        
    def setup_interpolating_grids(self):
        '''setup the grid to interpolate to, trim to necessary size of source grid, and compute interpolation weights'''
        # Setup grid to interpolate to
        xx, yy, zz = np.meshgrid(self.newx,self.newy,self.newz, indexing = 'xy')
        self.newgrid = np.ndarray(shape=(np.size(np.reshape(xx,-1)),3), dtype=np.float32)
        self.newgrid[:,0] = np.reshape(xx,-1)
        self.newgrid[:,1] = np.reshape(yy,-1)
        self.newgrid[:,2] = np.reshape(zz,-1)
        self.plots[self.plottype]['newgrid'] = self.newgrid

        return
    
    def setup_octree(self):
        tic=time.time()
        N_4=int(self.Ncell/4)
        # initialize cffi arrays with Python lists may be slightly faster
        x=OCTREE_BLOCK_GRID_FFI.new("float[]",list(self.x)) 
        y=OCTREE_BLOCK_GRID_FFI.new("float[]",list(self.y))
        z=OCTREE_BLOCK_GRID_FFI.new("float[]",list(self.z))
        self.xr_blk=OCTREE_BLOCK_GRID_FFI.new("float[]",N_4) # 2 elemts per block, smallest blocks have 2x2x2=8 positions
        self.yr_blk=OCTREE_BLOCK_GRID_FFI.new("float[]",N_4) # at cell corners (GUMICS), so maximum array size is N/4
        self.zr_blk=OCTREE_BLOCK_GRID_FFI.new("float[]",N_4) # BATSRUS blocks have at least 4x4x4 cell centers
        self.x_blk=OCTREE_BLOCK_GRID_FFI.new("float[]",N_4) # 2 elements per block, smallest blocks have 2x2x2=8 positions
        self.y_blk=OCTREE_BLOCK_GRID_FFI.new("float[]",N_4) # at cell corners (GUMICS), so maximum array size is N/4
        self.z_blk=OCTREE_BLOCK_GRID_FFI.new("float[]",N_4) # BATSRUS blocks have at least 4x4x4 cell centers
        box_range=OCTREE_BLOCK_GRID_FFI.new("float[]",6)
        self.N_blks=N_blks=OCTREE_BLOCK_GRID_LIB.xyz_ranges(self.Ncell,x,y,z,self.xr_blk,self.yr_blk,self.zr_blk,self.x_blk,self.y_blk,self.z_blk,box_range,1)
        N_octree=int(N_blks*8/7)
        self.octree_blocklist=OCTREE_BLOCK_GRID_FFI.new("octree_block[]",N_octree)
        # calculate maximum AMR level from box size in X and the minimum size of a block in X:
        self.XMIN=XMIN=float(box_range[0])
        self.XMAX=XMAX=float(box_range[1])
        self.YMIN=YMIN=float(box_range[2])
        self.YMAX=YMAX=float(box_range[3])
        self.ZMIN=ZMIN=float(box_range[4])
        self.ZMAX=ZMAX=float(box_range[5])
        dx_blk=np.arange(0,N_blks,1)
        dy_blk=np.arange(0,N_blks,1)
        dz_blk=np.arange(0,N_blks,1)
        for i in range(0,N_blks):
            dx_blk[i]=self.xr_blk[2*i+1]-self.xr_blk[2*i]
            dy_blk[i]=self.yr_blk[2*i+1]-self.yr_blk[2*i]
            dz_blk[i]=self.zr_blk[2*i+1]-self.zr_blk[2*i]
    
        dxmin_blk=np.min(dx_blk)
        dxmax_blk=np.max(dx_blk)
        dymin_blk=np.min(dy_blk)
        dymax_blk=np.max(dy_blk)
        dzmin_blk=np.min(dz_blk)
        dzmax_blk=np.max(dz_blk)
        p1=int(np.floor((XMAX-XMIN)/dxmax_blk+0.5))
        p2=int(np.floor((YMAX-YMIN)/dymax_blk+0.5))
        p3=int(np.floor((ZMAX-ZMIN)/dzmax_blk+0.5))
        while (int(p1/2)*2 == p1) & (int(p2/2)*2 == p2) & (int(p3/2)*2 == p3):
            p1=int(p1/2)
            p2=int(p2/2)
            p3=int(p3/2)

        MAX_AMRLEVEL=int(np.log((XMAX-XMIN)/(p1*dxmin_blk))/np.log(2.)+0.5 )
        print("p1,p2,p3: ",p1,p2,p3," MAX_AMRLEVEL:",MAX_AMRLEVEL)
        self.numparents_at_AMRlevel=OCTREE_BLOCK_GRID_FFI.new("int[]",N_blks*(MAX_AMRLEVEL+1))
        self.block_at_AMRlevel=OCTREE_BLOCK_GRID_FFI.new("int[]",N_blks*(MAX_AMRLEVEL+1))
        success=int(-1)
        success=OCTREE_BLOCK_GRID_LIB.setup_octree(N_blks,self.xr_blk,self.yr_blk,self.zr_blk,MAX_AMRLEVEL,box_range,self.octree_blocklist,N_octree,self.numparents_at_AMRlevel,self.block_at_AMRlevel)
        toc=time.time()
        print('setup_octree: ',toc-tic,'seconds elapsed')

    def register_variable(self, varname, units):
        """register variables into Kamodo for this model, SWMF-GM"""
        def interpolator(xvec):
              tic=time.time()
              var_data=OCTREE_BLOCK_GRID_FFI.new("float[]",list(self.variables[varname]['data']))
              xvec=np.array(xvec)
              xpos=OCTREE_BLOCK_GRID_FFI.new("float[]",list(xvec[:,0]))
              ypos=OCTREE_BLOCK_GRID_FFI.new("float[]",list(xvec[:,1]))
              zpos=OCTREE_BLOCK_GRID_FFI.new("float[]",list(xvec[:,2]))
              npos=len(list(xvec[:,0]))
              return_data=list(np.zeros(npos,dtype=np.float))
              return_data_ffi=OCTREE_BLOCK_GRID_FFI.new("float[]",return_data)
 #             print(xpos,ypos,zpos,npos)
 #             print(var_data,return_data_ffi)
              python_loop=False
              if python_loop:
                  for i in range(npos):
                      return_data[i]=OCTREE_BLOCK_GRID_LIB.interpolate_amrdata(xvec[i,0],xvec[i,1],xvec[i,2],var_data,1)
              else:
                  IS_OK=OCTREE_BLOCK_GRID_LIB.interpolate_amrdata_multipos(xpos,ypos,zpos,npos,var_data,return_data_ffi)
                  return_data[:]=list(return_data_ffi)
              toc=time.time()
#              print(varname,' interpolator: ',toc-tic,'seconds elapsed')
              return return_data
              
#        interpolator = self.get_grid_interpolator(varname)
        
        # store the interpolator
        self.variables[varname]['interpolator'] = interpolator

#        def interpolate(xvec):  
#            # Reduce size of read data block to speed up the interpolation
#            tmpgrid2 = self.grid[(self.grid[:,0] > (np.amin(xvec[:,0])-self.tol)) & 
#                                 (self.grid[:,0] < (np.amax(xvec[:,0])+self.tol)) &
#                                 (self.grid[:,1] > (np.amin(xvec[:,1])-self.tol)) & 
#                                 (self.grid[:,1] < (np.amax(xvec[:,1])+self.tol)) & 
#                                 (self.grid[:,2] > (np.amin(xvec[:,2])-self.tol)) & 
#                                 (self.grid[:,2] < (np.amax(xvec[:,2])+self.tol))]
#            data =  self.variables[varname]['data']
#            tmpdata = data[(self.grid[:,0] > (np.amin(xvec[:,0])-self.tol)) & 
#                           (self.grid[:,0] < (np.amax(xvec[:,0])+self.tol)) &
#                           (self.grid[:,1] > (np.amin(xvec[:,1])-self.tol)) & 
#                           (self.grid[:,1] < (np.amax(xvec[:,1])+self.tol)) & 
#                           (self.grid[:,2] > (np.amin(xvec[:,2])-self.tol)) & 
#                           (self.grid[:,2] < (np.amax(xvec[:,2])+self.tol))]
#            # Precompute weights for interpolation to optimize
#            tmpvtx, tmpwts = self.GMcompute_weights(tmpgrid2, xvec)
#            return self.GMinterpolate(tmpdata, tmpvtx, tmpwts)

        
        # update docstring for this variable
#        interpolate.__doc__ = "A function that returns {} in [{}].".format(varname,units)

#        self[varname] = kamodofy(interpolate, 
        self[varname] = kamodofy(interpolator, 
                                 units = units, 
                                 citation = "Rastaetter 2021",
                                 data = None)
        #self[varname + '_ijk'] = kamodofy(gridify(self[varname], 
        #                                          x_i = self.newgrid[:,0], 
        #                                          y_j = self.newgrid[:,1], 
        #                                          z_k = self.newgrid[:,2]),
        #                                  units = units,
        #                                  citation = "De Zeeuw 2020",
        #                                  data = self.variables[varname]['data'])

#    def get_grid_interpolator(self, varname):
#        """create a grid interpolator for passed variable based on Delaunay triangulation of points"""
#        data =  self.variables[varname]['data']
#        data2 = data[(self.grid[:,0] > (np.amin(self.newx)-self.tol)) & 
#                     (self.grid[:,0] < (np.amax(self.newx)+self.tol)) &
#                     (self.grid[:,1] > (np.amin(self.newy)-self.tol)) & 
#                     (self.grid[:,1] < (np.amax(self.newy)+self.tol)) & 
#                     (self.grid[:,2] > (np.amin(self.newz)-self.tol)) & 
#                     (self.grid[:,2] < (np.amax(self.newz)+self.tol))]
#        interpolator = self.GMinterpolate(data2, self.vtx, self.wts)
#        # interpolator = spint.griddata(self.grid2, data2, self.newgrid, method='linear')
#
#        return interpolator
    
#    def reinterpolate_values(self):
#        '''recompute interpolation from saved grid and weights and store values for plotting'''
#        for varname in self.variables:
#            data =  self.variables[varname]['data']
#            data2 = data[(self.grid[:,0] > (np.amin(self.newx)-self.tol)) & 
#                         (self.grid[:,0] < (np.amax(self.newx)+self.tol)) &
#                         (self.grid[:,1] > (np.amin(self.newy)-self.tol)) & 
#                         (self.grid[:,1] < (np.amax(self.newy)+self.tol)) & 
#                         (self.grid[:,2] > (np.amin(self.newz)-self.tol)) & 
#                         (self.grid[:,2] < (np.amax(self.newz)+self.tol))]
#            self.variables[varname]['interpolator'] = self.GMinterpolate(data2, self.vtx, self.wts)
#
#        return

#    def GMcompute_weights(self, xyz, uvw):
#        """compute the weights to optimize griddata interpolation using a constructed Delaunay grid"""
#        d = 3
#        tri = qhull.Delaunay(xyz)
#        simplex = tri.find_simplex(uvw)
#        vertices = np.take(tri.simplices, simplex, axis=0)
#        temp = np.take(tri.transform, simplex, axis=0)
#        delta = uvw - temp[:, d]
#        bary = np.einsum('njk,nk->nj', temp[:, :d, :], delta)
#        return vertices, np.hstack((bary, 1 - bary.sum(axis=1, keepdims=True)))

#    def GMinterpolate(self, values, vtx, wts):
#        """
#        Use precomputed weights to complete the interpolation.
#        Note: Use of a Delaunay triangulation with a cartesian grid will result in any given point
#              having the possibility of 8 different point becoming the 4 used points for the tetrahedal.
#              Thus, the interpolation can randomly vary the interpolated value with any changes to the
#              number of points in the original grid.
#        """
#        return np.einsum('nj,nj->n', np.take(values, vtx), wts)

    def trace_fieldline(self,x0,y0,z0,vector,dn,max_step,tilt=None,bdp=None):
        # bi-directional fiel line tracer using teh C program trace_fieldline_octree and interpolate_amrdata modified for Python
        # Lutz Rastaetter, Nov. 30, 2021
        #
        #int trace_fieldline_octree(float x_start,float y_start,float z_start, float r_end,
        #			   IDL_LONG nf, 
        #                           float *bx, float *by, float *bz,
        #//IDL_LONG *v_indices,
        #                           float *flx, float *fly, float *flz, IDL_LONG *step_max,
        #                           float dn, float bdp, float *tilt, float spherical_deg2rad);
        from readers.OCTREE_BLOCK_GRID._interpolate_amrdata import ffi as OCTREE_BLOCK_GRID_FFI
        from readers.OCTREE_BLOCK_GRID._interpolate_amrdata import lib as OCTREE_BLOCK_GRID_LIB

        flx0=OCTREE_BLOCK_GRID_FFI.new("float[]",max_step)
        fly0=OCTREE_BLOCK_GRID_FFI.new("float[]",max_step)
        flz0=OCTREE_BLOCK_GRID_FFI.new("float[]",max_step)
        flx1=OCTREE_BLOCK_GRID_FFI.new("float[]",max_step)
        fly1=OCTREE_BLOCK_GRID_FFI.new("float[]",max_step)
        flz1=OCTREE_BLOCK_GRID_FFI.new("float[]",max_step)

        if vector is None:
            vector=['bx','by','bz']
    
        bx=OCTREE_BLOCK_GRID_FFI.new("float[]",list(self.variables[vector[0]]['data']))
        by=OCTREE_BLOCK_GRID_FFI.new("float[]",list(self.variables[vector[1]]['data']))
        bz=OCTREE_BLOCK_GRID_FFI.new("float[]",list(self.variables[vector[2]]['data']))
            
        step_max0=OCTREE_BLOCK_GRID_FFI.new("int[]",[max_step])
        if tilt is None:
            tilt=OCTREE_BLOCK_GRID_FFI.new("float[]",[0.,0.])
        if bdp is None:
            bdp=0. # nonzero only if using perturbation field bx1,by1,bz1 
        if dn is None:
            dn=0.2 # iteration step as fraction of cell size
        # forward tracing
        status0=OCTREE_BLOCK_GRID_LIB.trace_fieldline_octree(x0,y0,z0,self.meta['R'],
                                                             bx,by,bz,flx0,fly0,flz0,step_max0,
                                                             dn,bdp,tilt,0.)
        # backward tracing
        step_max1=OCTREE_BLOCK_GRID_FFI.new("int[]",[max_step])
        status1=OCTREE_BLOCK_GRID_LIB.trace_fieldline_octree(x0,y0,z0,self.meta['R'],
                                                             bx,by,bz,flx1,fly1,flz1,step_max1,
                                                             -dn,bdp,tilt,0.)
        # return values: number of vertices and vertex positions (x,y,z)
        step_max=step_max0[0]+step_max1[0]-1
        flx=np.zeros(step_max)
        fly=np.zeros(step_max)
        flz=np.zeros(step_max)
        s=np.zeros(step_max)

        for i in range(step_max0[0]):
            flx[step_max1[0]-1+i]=flx0[i]
            fly[step_max1[0]-1+i]=fly0[i]
            flz[step_max1[0]-1+i]=flz0[i]
            s[step_max1[0]-1+i]=dn*i
        for i in range(step_max1[0]):
            flx[step_max1[0]-1-i]=flx1[i]
            fly[step_max1[0]-1-i]=fly1[i]
            flz[step_max1[0]-1-i]=flz1[i]
            s[step_max1[0]-1-i]=-dn*i

        return [step_max,flx,fly,flz,s]

    def chk_plot(self):
        '''Check current values of plotting variables and print them out.'''
        print('Values set for plotting:')
        print(' plottype = ',self.plottype)
        print(' cut = ',self.cut)
        print(' cutV = ',self.cutV)
        print(' nX',self.nX)
        print(' nY',self.nY)
        print(' nZ',self.nZ)
        print(' newx',self.newx)
        print(' newy',self.newy)
        print(' newz',self.newz)
        print(' tol',self.tol)
        print(' full grid shape: ',self.grid.shape)
        print(' reduced size grid shape: ',self.grid2.shape)
        print(' interpolation grid shape: ',self.newgrid.shape)
        return
    
    def set_plot(self, plottype = "XY", cutV = 0.):
        '''Set plotting variables for available preset plot types: XY, YZ, YZ, XYZ'''
        tic = time.perf_counter()
        if plottype == self.plottype:
            if cutV == self.cutV:
                print('Plottype (',plottype,') and cut value (',cutV,') are unchanged, returning.')
                return
        
        if plottype == "XY":
            self.plottype = plottype
            self.cut = 'Z'
            self.cutV = cutV
            self.nX = 141
            self.nY = 81
            self.nZ = 1
            self.newx = np.linspace(-50., 20., self.nX)
            self.newy = np.linspace(-20., 20., self.nY)
            self.newz = cutV
            xx, yy, zz = np.meshgrid(self.newx,self.newy,self.newz, indexing = 'xy')
            self.newgrid = np.ndarray(shape=(np.size(np.reshape(xx,-1)),3), dtype=np.float32)
            self.newgrid[:,0]=np.reshape(xx,-1)
            self.newgrid[:,1]=np.reshape(yy,-1)
            self.newgrid[:,2]=np.reshape(zz,-1)
        elif plottype == "XZ":
            self.plottype = plottype
            self.cut = 'Y'
            self.cutV = cutV
            self.nX = 141
            self.nY = 1
            self.nZ = 81
            self.newx = np.linspace(-50., 20., self.nX)
            self.newy = cutV
            self.newz = np.linspace(-20., 20., self.nZ)
            xx, yy, zz = np.meshgrid(self.newx,self.newy,self.newz, indexing = 'xy')
            self.newgrid = np.ndarray(shape=(np.size(np.reshape(xx,-1)),3), dtype=np.float32)
            self.newgrid[:,0]=np.reshape(xx,-1)
            self.newgrid[:,1]=np.reshape(yy,-1)
            self.newgrid[:,2]=np.reshape(zz,-1)
        elif plottype == "YZ":
            self.plottype = plottype
            self.cut = 'X'
            self.cutV = cutV
            self.nX = 1
            self.nY = 141
            self.nZ = 81
            self.newx = cutV
            self.newy = np.linspace(-35., 35., self.nY)
            self.newz = np.linspace(-20., 20., self.nZ)
            xx, yy, zz = np.meshgrid(self.newx,self.newy,self.newz, indexing = 'xy')
            self.newgrid = np.ndarray(shape=(np.size(np.reshape(xx,-1)),3), dtype=np.float32)
            self.newgrid[:,0]=np.reshape(xx,-1)
            self.newgrid[:,1]=np.reshape(yy,-1)
            self.newgrid[:,2]=np.reshape(zz,-1)
        elif plottype == "XYZ3D": # for running tests against C interpolation code
            self.plottype = plottype
            self.cut = ''
            self.cutV = None
            self.nX = 101
            self.nY = 91
            self.nZ = 81
            self.newx = np.linspace(-50., 20., self.nX)
            self.newy = np.linspace(-15., 15., self.nY)
            self.newz = np.linspace(-10., 10., self.nZ)
            xx, yy, zz = np.meshgrid(self.newx,self.newy,self.newz, indexing = 'xy')
            self.newx = np.linspace(-50., 20., self.nX)
            self.newy = np.linspace(-15., 15., self.nY)
            self.newz = np.linspace(-10., 10., self.nZ)
            self.newgrid = np.ndarray(shape=(np.size(np.reshape(xx,-1)),3), dtype=np.float32)
            self.newgrid[:,0]=np.reshape(xx,-1)
            self.newgrid[:,1]=np.reshape(yy,-1)
            self.newgrid[:,2]=np.reshape(zz,-1)
        elif plottype == "XYZ":
            self.plottype = plottype
            return
        else:
            print('Error, unknown plottype. ',plottype)
            return

        self.plots[plottype] = dict(cut=self.cut, cutV=self.cutV, tol=self.tol,
                                    nX=self.nX, nY=self.nY, nZ=self.nZ,
                                    newx=self.newx, newy=self.newy, newz=self.newz)
        self.setup_interpolating_grids()
#        self.reinterpolate_values()

        for varname in self.variables:
            self.plots[plottype][varname]=self.variables[varname]['interpolator']

        toc = time.perf_counter()
        print(f"Time resetting plot and precomputing interpolations: {toc - tic:0.4f} seconds")
        return
    
    def get_plot(self, var, colorscale="Viridis", sym="F", vmin="", vmax=""):
        '''
        Return a plotly figure object for the available plot types set in set_plot()..
        colorscale = Viridis [default], Cividis, Rainbow, or BlueRed
        sym = F [default] for symetric colorscale around 0
        vmin, vmax: set minimum and maximum value for contour values, empty is actual min/max
        '''
        #Set some text strings
        txtbot="Model: BATSRUS,  Run: " + str(self.runname) + ",  " + str(self.gridSize) + " cells,  minimum dx=" + str(self.gridMinDx)
        txtbar=var + " [" + self.variables[var]['units'] + "]"
        
        # Get values from interpolation already computed
        result=np.array(self.variables[var]['interpolator'](self.newgrid))
        r = np.sqrt(np.square(self.newgrid[:,0]) + np.square(self.newgrid[:,1]) + np.square(self.newgrid[:,2]))
        if sym == "T":
            cmax = np.max(np.absolute(result[r > 2.999]))
            if vmax != "":
                cmax = abs(float(vmax))
            if vmin != "":
                cmax = max(cmax,abs(float(vmin)))
            cmin = -cmax
        else:
            cmin=np.amin(result[r > 2.999])
            cmax=np.amax(result[r > 2.999])
            if vmax != "":
                cmax = float(vmax)
            if vmin != "":
                cmin = float(vmin)
        
        if self.plottype == "XY":
            txttop="Z=" + str(self.plots[self.plottype]['cutV']) + " slice,  Time = " + self.filetime
            xint = self.newx
            yint = self.newy
            # Reshape interpolated values into 2D
            result2=np.reshape(result,(self.nY,self.nX))
            def plot_XY(xint = xint, yint = yint):
                return result2
            plotXY = Kamodo(plot_XY = plot_XY)
            fig = plotXY.plot(plot_XY = dict())
            fig.update_xaxes(title_text="",scaleanchor='y')
            fig.update_yaxes(title_text="Y [RE]")
            # Choose colorscale
            if colorscale == "Rainbow":
                fig.update_traces(
                    colorscale=[[0.00, 'rgb(0,0,255)'],
                                [0.25, 'rgb(0,255,255)'],
                                [0.50, 'rgb(0,255,0)'],
                                [0.75, 'rgb(255,255,0)'],
                                [1.00, 'rgb(255,0,0)']]
                )
            elif colorscale == "Cividis":
                fig.update_traces(colorscale="Cividis")
            elif colorscale == "BlueRed":
                fig.update_traces(colorscale="RdBu", reversescale=True)
            else:
                fig.update_traces(colorscale="Viridis")
            fig.update_traces(
                zmin=cmin, zmax=cmax,
                ncontours=201,
                colorbar=dict(title=txtbar),
                hovertemplate="X: %{x:.2f}<br>Y: %{y:.2f}<br><b> %{z:.4f}</b><extra></extra>",
                contours=dict(coloring="fill", showlines=False)
            )
            fig.update_layout(
                title_text=txttop,
                annotations=[
                    dict(text="X [RE]", x=0.5, y=-0.10, showarrow=False, xref="paper", yref="paper", font=dict(size=14)),
                    dict(text=txtbot,
                         font=dict(size=16, family="sans serif", color="#000000"),
                         x=0.0, y=0.0, ax=0, ay=0, xanchor="left", xshift=-65, yshift=-42, xref="paper", yref="paper"
                    )
                ]
            )
            if self.plots[self.plottype]['cutV'] == 0.:
                fig.update_layout(
                    shapes=[
                        dict(type="circle", xref="x", yref="y", x0=-3, y0=-3, x1=3, y1=3, fillcolor="black", line_color="black"),
                        dict(type="circle", xref="x", yref="y", x0=-1, y0=-1, x1=1, y1=1, fillcolor="black", line_color="white"),
                        dict(type="path", path= self.ellipse_arc(N=30), fillcolor="white", line_color="white")
                    ]
                )
            return fig
        
        if self.plottype == "XZ":
            txttop="Y=" + str(self.plots[self.plottype]['cutV']) + " slice,  Time = " + self.filetime
            xint = self.newx
            zint = self.newz
            # Reshape interpolated values into 2D
            result2=np.reshape(result,(self.nX,self.nZ))
            def plot_XZ(xint = xint, zint = zint):
                return result2
            plotXZ = Kamodo(plot_XZ = plot_XZ)
            fig = plotXZ.plot(plot_XZ = dict())
            fig.update_xaxes(title_text="",scaleanchor='y')
            fig.update_yaxes(title_text="Z [RE]")
            # Choose colorscale
            if colorscale == "Rainbow":
                fig.update_traces(
                    colorscale=[[0.00, 'rgb(0,0,255)'],
                                [0.25, 'rgb(0,255,255)'],
                                [0.50, 'rgb(0,255,0)'],
                                [0.75, 'rgb(255,255,0)'],
                                [1.00, 'rgb(255,0,0)']]
                )
            elif colorscale == "Cividis":
                fig.update_traces(colorscale="Cividis")
            elif colorscale == "BlueRed":
                fig.update_traces(colorscale="RdBu", reversescale=True)
            else:
                fig.update_traces(colorscale="Viridis")
            fig.update_traces(
                zmin=cmin, zmax=cmax,
                ncontours=201,
                colorbar=dict(title=txtbar),
                hovertemplate="X: %{x:.2f}<br>Z: %{y:.2f}<br><b> %{z:.4f}</b><extra></extra>",
                contours=dict(coloring="fill", showlines=False)
            )
            fig.update_layout(
                title_text=txttop,
                annotations=[
                    dict(text="X [RE]", x=0.5, y=-0.10, showarrow=False, xref="paper", yref="paper", font=dict(size=14)),
                    dict(text=txtbot,
                         font=dict(size=16, family="sans serif", color="#000000"),
                         x=0.0, y=0.0, ax=0, ay=0, xanchor="left", xshift=-65, yshift=-42, xref="paper", yref="paper"
                    )
                ]
            )
            if self.plots[self.plottype]['cutV'] == 0.:
                fig.update_layout(
                    shapes=[
                        dict(type="circle", xref="x", yref="y", x0=-3, y0=-3, x1=3, y1=3, fillcolor="black", line_color="black"),
                        dict(type="circle", xref="x", yref="y", x0=-1, y0=-1, x1=1, y1=1, fillcolor="black", line_color="white"),
                        dict(type="path", path= self.ellipse_arc(N=30), fillcolor="white", line_color="white")
                    ]
            )
            return fig
        
        if self.plottype == "YZ":
            txttop="X=" + str(self.plots[self.plottype]['cutV']) + " slice,  Time = " + self.filetime
            yint = self.newy
            zint = self.newz
            # Reshape interpolated values into 2D
            result2=np.reshape(result,(self.nY,self.nZ))
            def plot_YZ(yint = yint, zint = zint):
                return result2
            plotYZ = Kamodo(plot_YZ = plot_YZ)
            fig = plotYZ.plot(plot_YZ = dict())
            fig.update_xaxes(title_text="",scaleanchor='y')
            fig.update_yaxes(title_text="Z [RE]")
            # Choose colorscale
            if colorscale == "Rainbow":
                fig.update_traces(
                    colorscale=[[0.00, 'rgb(0,0,255)'],
                                [0.25, 'rgb(0,255,255)'],
                                [0.50, 'rgb(0,255,0)'],
                                [0.75, 'rgb(255,255,0)'],
                                [1.00, 'rgb(255,0,0)']]
                )
            elif colorscale == "Cividis":
                fig.update_traces(colorscale="Cividis")
            elif colorscale == "BlueRed":
                fig.update_traces(colorscale="RdBu", reversescale=True)
            else:
                fig.update_traces(colorscale="Viridis")
            fig.update_traces(
                zmin=cmin, zmax=cmax,
                ncontours=201,
                colorbar=dict(title=txtbar),
                hovertemplate="Y: %{x:.2f}<br>Z: %{y:.2f}<br><b> %{z:.4f}</b><extra></extra>",
                contours=dict(coloring="fill", showlines=False)
            )
            fig.update_layout(
                title_text=txttop,
                annotations=[
                    dict(text="Y [RE]", x=0.5, y=-0.10, showarrow=False, xref="paper", yref="paper", font=dict(size=14)),
                    dict(text=txtbot,
                         font=dict(size=16, family="sans serif", color="#000000"),
                         x=0.0, y=0.0, ax=0, ay=0, xanchor="left", xshift=-65, yshift=-42, xref="paper", yref="paper"
                    )
                ]
            )
            if self.plots[self.plottype]['cutV'] == 0.:
                fig.update_layout(
                    shapes=[
                        dict(type="circle", xref="x", yref="y", x0=-3, y0=-3, x1=3, y1=3, fillcolor="black", line_color="black"),
                        dict(type="circle", xref="x", yref="y", x0=-1, y0=-1, x1=1, y1=1, fillcolor="white", line_color="white"),
                        #dict(type="path", path= self.ellipse_arc(N=30), fillcolor="white", line_color="white")
                    ]
            )
            return fig
        
        if self.plottype == "XYZ":
            Nplot = 0
            txttop="3D Plot,  Time = " + self.filetime

            # Plot 'XY' is assumed to be part of the 3D plot
            pt = 'XY'
            # Get values from interpolation already computed
            # result=self.plots[pt][var]
            newgrid = self.plots[pt]['newgrid']
            result=np.array(self.variables[var]['interpolator'](newgrid))

            r = np.sqrt(np.square(newgrid[:,0]) + np.square(newgrid[:,1]) + np.square(newgrid[:,2]))
            if sym == "T":
                cmax = np.max(np.absolute(result[(r[:] > 2.999)]))
                if vmax != "":
                    cmax = abs(float(vmax))
                if vmin != "":
                    cmax = max(cmax,abs(float(vmin)))
                    cmin = -cmax
            else:
                cmin=np.amin(result[(r[:] > 2.999)])
                cmax=np.amax(result[(r[:] > 2.999)])
                if vmax != "":
                    cmax = float(vmax)
                if vmin != "":
                    cmin = float(vmin)
            xint = self.plots[pt]['newx']
            yint = self.plots[pt]['newy']
            zint = self.plots[pt]['newz']
            # Reshape interpolated values into 3D
            result2=np.reshape(result,(self.plots[pt]['nY'],self.plots[pt]['nX'],self.plots[pt]['nZ']))
            def plot_XY(xint = xint, yint = yint, zint = zint):
                return result2
            plotXY = Kamodo(plot_XY = plot_XY)
            fig = plotXY.plot(plot_XY = dict())
            Nplot = Nplot +1

            if('XZ' in self.plots):
                pt = 'XZ'
                # Get values from interpolation already computed
#                result=self.plots[pt][var]
                newgrid = self.plots[pt]['newgrid']
                result=np.array(self.variables[var]['interpolator'](newgrid))
                r = np.sqrt(np.square(newgrid[:,0]) + np.square(newgrid[:,1]) + np.square(newgrid[:,2]))
                if sym == "T":
                    tcmax = np.max(np.absolute(result[(r[:] > 2.999)]))
                    if vmax != "":
                        tcmax = abs(float(vmax))
                    if vmin != "":
                        tcmax = max(tcmax,abs(float(vmin)))
                    cmin = -tcmax
                else:
                    tcmin=np.amin(result[(r[:] > 2.999)])
                    tcmax=np.amax(result[(r[:] > 2.999)])
                    if vmax != "":
                        tcmax = float(vmax)
                    if vmin != "":
                        tcmin = float(vmin)
                cmin=min(cmin,tcmin)
                cmax=max(cmax,tcmax)
                xint = self.plots[pt]['newx']
                yint = self.plots[pt]['newy']
                zint = self.plots[pt]['newz']
                # Reshape interpolated values into 3D
                result2=np.reshape(result,(self.plots[pt]['nY'],self.plots[pt]['nX'],self.plots[pt]['nZ']))
                def plot_XZ(xint = xint, yint = yint, zint = zint):
                    return result2
                plotXZ = Kamodo(plot_XZ = plot_XZ)
                figXZ = plotXZ.plot(plot_XZ = dict())
                fig.add_surface(showscale=False)
                fig.data[Nplot].surfacecolor = figXZ.data[0].surfacecolor
                fig.data[Nplot].x = figXZ.data[0].x
                fig.data[Nplot].y = figXZ.data[0].y
                fig.data[Nplot].z = figXZ.data[0].z
                Nplot = Nplot +1

            if('YZ' in self.plots):
                pt = 'YZ'
                # Get values from interpolation already computed
                #result=self.plots[pt][var]
                newgrid = self.plots[pt]['newgrid']
                result=np.array(self.variables[var]['interpolator'](newgrid))
                r = np.sqrt(np.square(newgrid[:,0]) + np.square(newgrid[:,1]) + np.square(newgrid[:,2]))
                if sym == "T":
                    tcmax = np.max(np.absolute(result[(r[:] > 2.999)]))
                    if vmax != "":
                        tcmax = abs(float(vmax))
                    if vmin != "":
                        tcmax = max(tcmax,abs(float(vmin)))
                    cmin = -tcmax
                else:
                    tcmin=np.amin(result[(r[:] > 2.999)])
                    tcmax=np.amax(result[(r[:] > 2.999)])
                    if vmax != "":
                        tcmax = float(vmax)
                    if vmin != "":
                        tcmin = float(vmin)
                cmin=min(cmin,tcmin)
                cmax=max(cmax,tcmax)
                xint = self.plots[pt]['newx']
                yint = self.plots[pt]['newy']
                zint = self.plots[pt]['newz']
                # Reshape interpolated values into 3D
                result2=np.reshape(result,(self.plots[pt]['nY'],self.plots[pt]['nX'],self.plots[pt]['nZ']))
                def plot_YZ(xint = xint, yint = yint, zint = zint):
                    return result2
                plotYZ = Kamodo(plot_YZ = plot_YZ)
                figYZ = plotYZ.plot(plot_YZ = dict())
                fig.add_surface(showscale=False)
                fig.data[Nplot].surfacecolor = figYZ.data[0].surfacecolor
                fig.data[Nplot].x = figYZ.data[0].x
                fig.data[Nplot].y = figYZ.data[0].y
                fig.data[Nplot].z = figYZ.data[0].z
                Nplot = Nplot +1

            # Choose colorscale
            if colorscale == "Rainbow":
                fig.update_traces(
                    colorscale=[[0.00, 'rgb(0,0,255)'],
                                [0.25, 'rgb(0,255,255)'],
                                [0.50, 'rgb(0,255,0)'],
                                [0.75, 'rgb(255,255,0)'],
                                [1.00, 'rgb(255,0,0)']]
                )
            elif colorscale == "RdBu":
                fig.update_traces(colorscale="RdBu")
            elif colorscale == "Cividis":
                fig.update_traces(colorscale="Cividis")
            elif colorscale == "BlueRed":
                fig.update_traces(colorscale="RdBu", reversescale=True)
            else:
                fig.update_traces(colorscale="Viridis")
            fig.update_traces(
                colorbar=dict(title=txtbar),
                cmin=cmin, cmax=cmax,
                hovertemplate="X: %{x:.2f}<br>Y: %{y:.2f}<br>Z: %{z:.2f}<extra></extra>"
            )
            fig.update_layout(
                scene_aspectmode='data',
                title_text=txttop,
                scene=dict(xaxis=dict(title="X [RE]"),yaxis=dict(title="Y [RE]"),zaxis=dict(title="Z [RE]")),
                annotations=[
                    dict(text=txtbot,
                         font=dict(size=16, family="sans serif", color="#000000"),
                         x=-0.08, y=-0.05, ax=0, ay=0, xanchor="left", xref="paper", yref="paper"
                    )
                ],
                margin=dict(l=0, b=30, t=30)
            )

            # Add in plot axis for reference
            #fig.add_scatter3d()
            #fig.data[Nplot].x = [-20., 0., 20., 0.,   0., 0.,  0., 0.,   0., 0.,  0.]
            #fig.data[Nplot].y = [  0., 0., 0.,  0., -20., 0., 20., 0.,   0., 0.,  0.]
            #fig.data[Nplot].z = [  0., 0., 0.,  0.,   0., 0.,  0., 0., -20., 0., 20.]
            #fig.data[Nplot].name = 'Axis'
            #fig.data[Nplot].marker = dict(size=3, color='white', opacity=0.8)
            #fig.data[Nplot].line = dict(color='white', width=4)
            #fig.data[Nplot].hoverinfo='skip'
            #Nplot = Nplot +1

            # Add in R=3 sphere
            #dataXYZ = pd.read_csv('http://vmr.engin.umich.edu/dbg/sphereXYZ.csv')
            #dataIJK = pd.read_csv('http://vmr.engin.umich.edu/dbg/sphereIJK.csv')
            dataXYZ = pd.read_csv('https://ccmc.gsfc.nasa.gov/Kamodo/demo/sphereXYZ.csv')
            dataIJK = pd.read_csv('https://ccmc.gsfc.nasa.gov/Kamodo/demo/sphereIJK.csv')
            fig.add_mesh3d()
            fig.data[Nplot].x = dataXYZ['x']*3.
            fig.data[Nplot].y = dataXYZ['y']*3.
            fig.data[Nplot].z = dataXYZ['z']*3.
            fig.data[Nplot].i = dataIJK['i']
            fig.data[Nplot].j = dataIJK['j']
            fig.data[Nplot].k = dataIJK['k']
            fig.data[Nplot].facecolor = dataIJK['c']
            fig.data[Nplot].flatshading = True
            fig.data[Nplot].name = 'R=3 sphere'
            fig.data[Nplot].hovertemplate="R=3 sphere<extra></extra>"
            Nplot = Nplot +1

            return fig
        
        print("ERROR, unknown plottype =",plottype,", exiting without definition of figure.")
        return
    
    # function to draw a partial ellipse (for sun-lit half of earth)
    def ellipse_arc(self, x_center=0, y_center=0, a=1, b=1, start_angle=-np.pi/2, end_angle=np.pi/2, N=100, closed= False):
        '''small function to overlay a partial ellipse on the plot for sun-lit half of earth effect'''
        t = np.linspace(start_angle, end_angle, N)
        x = x_center + a*np.cos(t)
        y = y_center + b*np.sin(t)
        path = f'M {x[0]}, {y[0]}'
        for k in range(1, len(t)):
            path += f'L{x[k]}, {y[k]}'
        if closed:
            path += ' Z'
        return path
    
#
# Functions outside of the module that are useful.
#
def show_GM_files(runpath = "./", 
                  runname = "noname", 
                  start_time = "1000/01/01 00:00", 
                  filepath = "/IO2/3d*.out"):
    '''
    This function will show a list of available SWMF-GM plot files with simulated date/time for that file.
    The resulting files will be read from: runpath + runname + filepath (ie. ./noname/IO2/3d*.out)
    '''
    
    print('Routine prints simulation data/time for each simulation output and returns last file.\n',
         ' If year is 1000, then no start_time for the run was found\n and arguement should be added to call.')
    
    # Pull start_time from the DatabaseInfo file (if it exists), then add simulated time from filenames
    if path.isfile(runpath+runname+'/DatabaseInfo'):
        for line in open(runpath+runname+'/DatabaseInfo'):
            if "start_time" in line:
                start_time = line.split("#", 1)[0].strip()
                print('... found run start_time = ',start_time)
                break
    print('')
    files = glob.glob(runpath+runname+filepath)
    files.sort()
    for file in files:
        runtime = file.split("_t")[-1].strip()
        runtime = str(runtime.split("_")[0].strip())
        hour=int(runtime[0:4])
        minute=int(runtime[4:6])
        second=int(runtime[6:])
        dt = datetime.strptime(start_time, '%Y/%m/%d %H:%M')
        dt = dt + timedelta(hours=hour, minutes=minute, seconds=second)
        filetime = dt.strftime("%Y/%m/%d %H:%M:%S UT")
        print(filetime,'   ',file)
    print('')
    return file




# -*- coding: utf-8 -*-
"""
Created on Wed Sep 21 15:57:01 2022

@author: rringuet
files given have one hour of data, open at end (days since 1979-01-01 00:00:00)
lon and mlon need wrapping
lon is 0-360 (shape=144), mlon is -180 to 180 (shape=80)
lat (shape=96) and mlat (shape=97) include both poles
two pressure level variables:
    lev.shape = 130 at midpoints (hPa)
    ilev.shape = 131 at interfaces (hPa)
data structure order = (time, lev/ilev, lat/mlat, lon/mlon)
"""
# starts with variables from the h1 files. No h2 files in this set
model_varnames = {'co2vmr': ['mmr_CO2', 'co2 volume mixing ratio', 0,
                             'GDZ', 'sph', ['time'], ''],
                  'ch4vmr': ['mmr_CH4', 'ch4 volume mixing ratio', 0,
                             'GDZ', 'sph', ['time'], ''],
                  'n2ovmr': ['mmr_N2O', 'n2o volume mixing ratio', 0,
                             'GDZ', 'sph', ['time'], ''],
                  'f11vmr': ['mmr_F11', 'f11 volume mixing ratio', 0,
                             'GDZ', 'sph', ['time'], ''],
                  'f12vmr': ['mmr_F12', 'f12 volume mixing ratio', 0,
                             'GDZ', 'sph', ['time'], ''],
                  'sol_tsi': ['TSI', 'total solar irradiance', 0, 'GDZ', 'sph',
                              ['time'], 'W/m**2'],
                  'colat_crit1': ['Colat1', 'First co-latitude of electro-' +
                                  'potential critical angle', 0, 'GDZ', 'sph',
                                  ['time'],  'degrees'],
                  'colat_crit2': ['Colat2', 'Second co-latitude of electro-' +
                                  'potential critical angle', 0, 'GDZ', 'sph',
                                  ['time'], 'degrees'],
                  'ED1': ['E_east', 'ED1: Eastward Electric Field', 0,
                          'MAG', 'sph', ['time', 'mlon', 'mlat'], 'V/m'],
                  'ED2': ['E_equator', 'ED2: Equatorward Electric Field', 0,
                          'MAG', 'sph', ['time', 'mlon', 'mlat'], 'V/m'],
                  'EDYN_ZIGM11_PED': ['Sigma_P', 'Pedersen Conductance', 0,
                                      'MAG', 'sph', ['time', 'mlon', 'mlat'],
                                      'S'],
                  'EDYN_ZIGM2_HAL': ['Sigma_H', 'Hall Conductance', 0, 'MAG',
                                     'sph', ['time', 'mlon', 'mlat'], 'S'],
                  'ElecColDens': ['TEC', 'Electron Column Density', 0, 'GDZ',
                                  'sph', ['time', 'lon', 'lat'],
                                  '10**16/m**2'],
                  'H': ['Hyd_ilev', 'H concentration', 0, 'GDZ', 'sph',
                        ['time', 'lon', 'lat', 'ilev'], 'mol/mol'],
                  'H_2': ['Hyd', 'H concentration', 0, 'GDZ', 'sph',
                          ['time', 'lon', 'lat', 'height'], 'mol/mol'],
                  'O': ['Oxy_ilev', 'O concentration', 0, 'GDZ', 'sph',
                        ['time', 'lon', 'lat', 'ilev'], 'mol/mol'],
                  'O_2': ['Oxy', 'O concentration', 0, 'GDZ', 'sph',
                          ['time', 'lon', 'lat', 'height'], 'mol/mol'],
                  'O2': ['O2_ilev', 'O2 concentration', 0, 'GDZ', 'sph',
                         ['time', 'lon', 'lat', 'ilev'], 'mol/mol'],
                  'O2_2': ['O2', 'O2 concentration', 0, 'GDZ', 'sph',
                           ['time', 'lon', 'lat', 'height'], 'mol/mol'],
                  'OMEGA': ['v_up_ilev', 'Vertical velocity (pressure)', 0,
                            'GDZ', 'sph', ['time', 'lon', 'lat', 'ilev'],
                            'Pa/s'],
                  'OMEGA_2': ['v_up', 'Vertical velocity (pressure)', 0, 'GDZ',
                              'sph', ['time', 'lon', 'lat', 'height'], 'Pa/s'],
                  'PHIM2D': ['phi', 'PHIM2D: Electric Potential', 0, 'MAG',
                             'sph', ['time', 'mlon', 'mlat'], 'V'],
                  'PS': ['P_surface', 'Surface pressure', 0, 'GDZ', 'sph',
                         ['time', 'lon', 'lat'], 'Pa'],
                  'T': ['T_ilev', 'Temperature', 0, 'GDZ', 'sph',
                        ['time', 'lon', 'lat', 'ilev'], 'K'],
                  'T_2': ['T', 'Temperature', 0, 'GDZ', 'sph',
                          ['time', 'lon', 'lat', 'height'], 'K'],
                  'TElec': ['T_e_ilev', 'Electron Temperature', 0, 'GDZ',
                            'sph', ['time', 'lon', 'lat', 'ilev'], 'K'],
                  'TElec_2': ['T_e', 'Electron Temperature', 0, 'GDZ', 'sph',
                              ['time', 'lon', 'lat', 'height'], 'K'],
                  'TIon': ['T_i_ilev', 'Ion Temperature', 0, 'GDZ', 'sph',
                           ['time', 'lon', 'lat', 'ilev'], 'K'],
                  'TIon_2': ['T_i', 'Ion Temperature', 0, 'GDZ', 'sph',
                             ['time', 'lon', 'lat', 'height'], 'K'],
                  'U': ['v_east_ilev', 'Zonal wind', 0, 'GDZ', 'sph',
                        ['time', 'lon', 'lat', 'ilev'], 'm/s'],
                  'U_2': ['v_east', 'Zonal wind', 0, 'GDZ', 'sph',
                          ['time', 'lon', 'lat', 'height'], 'm/s'],
                  'UI': ['vi_east_ilev', 'UI Zonal ion drift from edynamo', 0,
                         'GDZ', 'sph', ['time', 'lon', 'lat', 'ilev'], 'm/s'],
                  'UI_2': ['vi_east', 'UI Zonal ion drift from edynamo', 0,
                           'GDZ', 'sph', ['time', 'lon', 'lat', 'height'],
                           'm/s'],
                  'V': ['v_north_ilev', 'Meridional wind', 0, 'GDZ', 'sph',
                        ['time', 'lon', 'lat', 'ilev'], 'm/s'],
                  'V_2': ['v_north', 'Meridional wind', 0, 'GDZ', 'sph',
                          ['time', 'lon', 'lat', 'height'], 'm/s'],
                  'VI': ['vi_north_ilev', 'VI Meridional ion drift from ' +
                         'edynamo', 0, 'GDZ', 'sph',
                         ['time', 'lon', 'lat', 'ilev'], 'm/s'],
                  'VI_2': ['vi_north', 'VI Meridional ion drift from edynamo',
                           0, 'GDZ', 'sph', ['time', 'lon', 'lat', 'height'],
                           'm/s'],
                  'WI': ['vi_up_ilev', 'WI Vertical ion drift from edynamo',
                         0, 'GDZ', 'sph', ['time', 'lon', 'lat', 'ilev'],
                         'm/s'],
                  'WI_2': ['vi_up', 'WI Vertical ion drift from edynamo',
                           0, 'GDZ', 'sph', ['time', 'lon', 'lat', 'height'],
                           'm/s'],
                  'Z3': ['H_geopot_ilev', 'Geopotential Height (above sea ' +
                         'level)', 0, 'GDZ', 'sph',
                         ['time', 'lon', 'lat', 'ilev'], 'm'],
                  'e': ['e_ilev', 'e concentration', 0, 'GDZ', 'sph',
                        ['time', 'lon', 'lat', 'ilev'], 'mol/mol'],
                  'e_2': ['e', 'e concentration', 0, 'GDZ', 'sph',
                          ['time', 'lon', 'lat', 'height'], 'mol/mol'],
                  'RHO_CLUBB': ['rho_ilev1', 'Air density', 0, 'GDZ', 'sph',
                                ['time', 'lon', 'lat', 'ilev1'], 'kg/m**3'],
                  'RHO_CLUBB_lev': ['rho_ilev', 'Air density', 0, 'GDZ', 'sph',
                                    ['time', 'lon', 'lat', 'ilev'], 'kg/m**3'],
                  'RHO_CLUBB_lev_2': ['rho', 'Air density', 0, 'GDZ', 'sph',
                                      ['time', 'lon', 'lat', 'height'],
                                      'kg/m**3'],

                  # variables in h1 files also in h3 files
                  'Z3GM': ['H_geomet_ilev', 'Geometric height', 0, 'GDZ',
                           'sph', ['time', 'lon', 'lat', 'ilev'], 'm'],

                  # variables from the h2 files (also in h3 files!)
                  'EDens': ['N_e_ilev', 'e Number Density (sum of O2+,NO+' +
                            ',N2+,O+)', 0, 'GDZ', 'sph',
                            ['time', 'lon', 'lat', 'ilev'], 'cm**3'],
                  'EDens_2': ['N_e', 'e Number Density (sum of O2+,NO+,N2+' +
                              ',O+)', 0, 'GDZ', 'sph',
                              ['time', 'lon', 'lat', 'height'], 'cm**3'],
                  'HMF2': ['HmF2', 'Height of the F2 Layer', 0, 'GDZ', 'sph',
                           ['time', 'lon', 'lat'], 'km'],
                  'NMF2': ['NmF2', 'Peak Density of the F2 Layer', 0, 'GDZ',
                           'sph', ['time', 'lon', 'lat'], '1/cm**3'],
                  'OPLUS': ['N_Oplus_ilev', 'O+ number density', 0, 'GDZ',
                            'sph', ['time', 'lon', 'lat', 'ilev'], '1/cm**3'],
                  'OPLUS_2': ['N_Oplus', 'O+ number density', 0, 'GDZ', 'sph',
                              ['time', 'lon', 'lat', 'height'], '1/cm**3'],

                  # variables from the h3 files
                  'CO2': ['CO2_ilev', 'CO2 concentration', 0, 'GDZ', 'sph',
                          ['time', 'lon', 'lat', 'ilev'], 'mol/mol'],
                  'CO2_2': ['CO2', 'CO2 concentration', 0, 'GDZ', 'sph',
                            ['time', 'lon', 'lat', 'height'], 'mol/mol'],
                  # 'EKGW': ['EKGW_ilev1', 'Effective Kzz due to diffusion by '
                  # gravity waves', 0, 'GDZ', 'sph',
                  #           ['time', 'lon', 'lat', 'ilev1'], 'm**2/s'],
                  # 'EKGW_2': ['EKGW', 'Effective Kzz due to diffusion by grav'
                  # ity waves', 0, 'GDZ', 'sph',
                  #          ['time', 'lon', 'lat', 'height'], 'm**2/s'],
                  'N': ['Nit_ilev', 'N concentration', 0, 'GDZ', 'sph',
                        ['time', 'lon', 'lat', 'ilev'], 'mol/mol'],
                  'N_2': ['Nit', 'N concentration', 0, 'GDZ', 'sph',
                          ['time', 'lon', 'lat', 'height'], 'mol/mol'],
                  'NO': ['NO_ilev', 'NO concentration', 0, 'GDZ', 'sph',
                         ['time', 'lon', 'lat', 'ilev'], 'mol/mol'],
                  'NO_2': ['NO', 'NO concentration', 0, 'GDZ', 'sph',
                           ['time', 'lon', 'lat', 'height'], 'mol/mol'],
                  'OpDens': ['N_Oplus_ilev', 'O+ Number Density', 0, 'GDZ',
                             'sph', ['time', 'lon', 'lat', 'ilev'], '1/cm**3'],
                  'OpDens_2': ['N_Oplus', 'O+ Number Density', 0, 'GDZ', 'sph',
                               ['time', 'lon', 'lat', 'height'], '1/cm**3'],
                  'QCO2': ['Q_CO2_ilev', 'CO2 cooling', 0, 'GDZ', 'sph',
                           ['time', 'lon', 'lat', 'ilev'], 'K/s'],
                  'QCO2_2': ['Q_CO2', 'CO2 cooling', 0, 'GDZ', 'sph',
                             ['time', 'lon', 'lat', 'height'], 'K/s'],
                  'QHC2S': ['Q_HC2S_ilev', 'Cooling to Space', 0, 'GDZ', 'sph',
                            ['time', 'lon', 'lat', 'ilev'], 'K/s'],
                  'QHC2S_2': ['Q_HC2S', 'Cooling to Space', 0, 'GDZ', 'sph',
                              ['time', 'lon', 'lat', 'height'], 'K/s'],
                  'QJOULE': ['Q_Joule_ilev', 'Joule Heat', 0, 'GDZ', 'sph',
                             ['time', 'lon', 'lat', 'ilev'], 'K/s'],
                  'QJOULE_2': ['Q_Joule', 'Joule Heat', 0, 'GDZ', 'sph',
                               ['time', 'lon', 'lat', 'height'], 'K/s'],
                  'QNO': ['Q_NO_ilev', 'NO cooling', 0, 'GDZ', 'sph',
                          ['time', 'lon', 'lat', 'ilev'], 'K/s'],
                  'QNO_2': ['Q_NO', 'NO cooling', 0, 'GDZ', 'sph',
                            ['time', 'lon', 'lat', 'height'], 'K/s'],
                  'QO3': ['Q_O3_ilev', 'O3 cooling', 0, 'GDZ', 'sph',
                          ['time', 'lon', 'lat', 'ilev'], 'K/s'],
                  'QO3_2': ['Q_O3', 'O3 cooling', 0, 'GDZ', 'sph',
                            ['time', 'lon', 'lat', 'height'], 'K/s'],
                  'QO3P': ['Q_O3P_ilev', 'O3P cooling', 0, 'GDZ', 'sph',
                           ['time', 'lon', 'lat', 'ilev'], 'K/s'],
                  'QO3P_2': ['Q_O3P', 'O3P cooling', 0, 'GDZ', 'sph',
                             ['time', 'lon', 'lat', 'height'], 'K/s'],
                  'QRS_TOT': ['Q_Total_ilev', 'Merged SW heating: QRS+QCP+' +
                              'QRS_EUV+QRS_CO2NIR+QRS_AUR+QTHERMAL', 0, 'GDZ',
                              'sph', ['time', 'lon', 'lat', 'ilev'], 'K/s'],
                  'QRS_TOT_2': ['Q_Total', 'Merged SW heating: QRS+QCP+' +
                                'QRS_EUV+QRS_CO2NIR+QRS_AUR+QTHERMAL', 0,
                                'GDZ', 'sph', ['time', 'lon', 'lat', 'height'],
                                'K/s'],
                  'SolIonRate_Tot': ['SolIonRate_Tot_ilev', 'reaction rate ' +
                                     'group', 0, 'GDZ', 'sph',
                                     ['time', 'lon', 'lat', 'ilev'],
                                     '1/cm**3/s'],  # molecules/cm**3/s
                  'SolIonRate_Tot_2': ['SolIonRate_Tot', 'reaction rate group',
                                       0, 'GDZ', 'sph',
                                       ['time', 'lon', 'lat', 'height'],
                                       '1/cm**3/s'],  # molecules/cm**3/s
                  'TTGW': ['TTGW_ilev', 'T tendency - gravity wave drag', 0,
                           'GDZ', 'sph', ['time', 'lon', 'lat', 'ilev'],
                           'K/s'],
                  'TTGW_2': ['TTGW', 'T tendency - gravity wave drag', 0,
                             'GDZ', 'sph', ['time', 'lon', 'lat', 'height'],
                             'K/s'],
                  'UTGW_TOTAL': ['UTGW_TOTAL_ilev', 'Total U tendency due to' +
                                 ' gravity wave drag', 0, 'GDZ', 'sph',
                                 ['time', 'lon', 'lat', 'ilev'], 'm/s**2'],
                  'UTGW_TOTAL_2': ['UTGW_TOTAL', 'Total U tendency due to ' +
                                   'gravity wave drag', 0, 'GDZ', 'sph',
                                   ['time', 'lon', 'lat', 'height'], 'm/s**2'],

                  # variables from the h4 files
                  'OMEGA_08_COS': ['OMEGA_08_COS_ilev', 'vertical pressure ' +
                                   'velocity  8hr. cos coeff.', 0, 'GDZ',
                                   'sph', ['time', 'lon', 'lat', 'ilev'],
                                   'Pa/s'],
                  'OMEGA_08_COS_2': ['OMEGA_08_COS', 'vertical pressure ' +
                                     'velocity  8hr. cos coeff.', 0, 'GDZ',
                                     'sph', ['time', 'lon', 'lat', 'height'],
                                     'Pa/s'],
                  'OMEGA_08_SIN': ['OMEGA_08_SIN_ilev', 'vertical pressure ' +
                                   'velocity  8hr. sin coeff.', 0, 'GDZ',
                                   'sph', ['time', 'lon', 'lat', 'ilev'],
                                   'Pa/s'],
                  'OMEGA_08_SIN_2': ['OMEGA_08_SIN', 'vertical pressure ' +
                                     'velocity  8hr. sin coeff.', 0, 'GDZ',
                                     'sph', ['time', 'lon', 'lat', 'height'],
                                     'Pa/s'],
                  'OMEGA_12_COS': ['OMEGA_12_COS_ilev', 'vertical pressure ' +
                                   'velocity 12hr. cos coeff.', 0, 'GDZ',
                                   'sph', ['time', 'lon', 'lat', 'ilev'],
                                   'Pa/s'],
                  'OMEGA_12_COS_2': ['OMEGA_12_COS', 'vertical pressure ' +
                                     'velocity 12hr. cos coeff.', 0, 'GDZ',
                                     'sph', ['time', 'lon', 'lat', 'height'],
                                     'Pa/s'],
                  'OMEGA_12_SIN': ['OMEGA_12_SIN_ilev', 'vertical pressure ' +
                                   'velocity 12hr. sin coeff.', 0, 'GDZ',
                                   'sph', ['time', 'lon', 'lat', 'ilev'],
                                   'Pa/s'],
                  'OMEGA_12_SIN_2': ['OMEGA_12_SIN', 'vertical pressure ' +
                                     'velocity 12hr. sin coeff.', 0, 'GDZ',
                                     'sph', ['time', 'lon', 'lat', 'height'],
                                     'Pa/s'],
                  'OMEGA_24_COS': ['OMEGA_24_COS_ilev', 'vertical pressure ' +
                                   'velocity 24hr. cos coeff.', 0, 'GDZ',
                                   'sph', ['time', 'lon', 'lat', 'ilev'],
                                   'Pa/s'],
                  'OMEGA_24_COS_2': ['OMEGA_24_COS', 'vertical pressure ' +
                                     'velocity 24hr. cos coeff.', 0, 'GDZ',
                                     'sph', ['time', 'lon', 'lat', 'height'],
                                     'Pa/s'],
                  'OMEGA_24_SIN': ['OMEGA_24_SIN_ilev', 'vertical pressure ' +
                                   'velocity 24hr. sin coeff.', 0, 'GDZ',
                                   'sph', ['time', 'lon', 'lat', 'ilev'],
                                   'Pa/s'],
                  'OMEGA_24_SIN_2': ['OMEGA_24_SIN', 'vertical pressure ' +
                                     'velocity 24hr. sin coeff.', 0, 'GDZ',
                                     'sph', ['time', 'lon', 'lat', 'height'],
                                     'Pa/s'],
                  'T_08_COS': ['T_08_COS_ilev', 'Temperature  8hr. cos coeff.',
                               0, 'GDZ', 'sph', ['time', 'lon', 'lat', 'ilev'],
                               'K'],
                  'T_08_COS_2': ['T_08_COS', 'Temperature  8hr. cos coeff.', 0,
                                 'GDZ', 'sph', ['time', 'lon', 'lat',
                                                'height'], 'K'],
                  'T_08_SIN': ['T_08_SIN_ilev', 'Temperature  8hr. sin coeff.',
                               0, 'GDZ', 'sph', ['time', 'lon', 'lat', 'ilev'],
                               'K'],
                  'T_08_SIN_2': ['T_08_SIN', 'Temperature  8hr. sin coeff.',
                                 0, 'GDZ', 'sph', ['time', 'lon', 'lat',
                                                   'height'], 'K'],
                  'T_12_COS': ['T_12_COS_ilev', 'Temperature 12hr. cos coeff.',
                               0, 'GDZ', 'sph', ['time', 'lon', 'lat', 'ilev'],
                               'K'],
                  'T_12_COS_2': ['T_12_COS', 'Temperature 12hr. cos coeff.',
                                 0, 'GDZ', 'sph', ['time', 'lon', 'lat',
                                                   'height'], 'K'],
                  'T_12_SIN': ['T_12_SIN_ilev', 'Temperature 12hr. sin coeff.',
                               0, 'GDZ', 'sph', ['time', 'lon', 'lat', 'ilev'],
                               'K'],
                  'T_12_SIN_2': ['T_12_SIN', 'Temperature 12hr. sin coeff.',
                                 0, 'GDZ', 'sph', ['time', 'lon', 'lat',
                                                   'height'], 'K'],
                  'T_24_COS': ['T_24_COS_ilev', 'Temperature 24hr. cos coeff.',
                               0, 'GDZ', 'sph', ['time', 'lon', 'lat', 'ilev'],
                               'K'],
                  'T_24_COS_2': ['T_24_COS', 'Temperature 24hr. cos coeff.',
                                 0, 'GDZ', 'sph', ['time', 'lon', 'lat',
                                                   'height'], 'K'],
                  'T_24_SIN': ['T_24_SIN_ilev', 'Temperature 24hr. sin coeff.',
                               0, 'GDZ', 'sph', ['time', 'lon', 'lat', 'ilev'],
                               'K'],
                  'T_24_SIN_2': ['T_24_SIN', 'Temperature 24hr. sin coeff.',
                                 0, 'GDZ', 'sph', ['time', 'lon', 'lat',
                                                   'height'], 'K'],
                  'U_08_COS': ['U_08_COS_ilev', 'Zonal wind  8hr. cos coeff.',
                               0, 'GDZ', 'sph', ['time', 'lon', 'lat', 'ilev'],
                               'm/s'],
                  'U_08_COS_2': ['U_08_COS', 'Zonal wind  8hr. cos coeff.',
                                 0, 'GDZ', 'sph', ['time', 'lon', 'lat',
                                                   'height'], 'm/s'],
                  'U_08_SIN': ['U_08_SIN_ilev', 'Zonal wind  8hr. sin coeff.',
                               0, 'GDZ', 'sph', ['time', 'lon', 'lat', 'ilev'],
                               'm/s'],
                  'U_08_SIN_2': ['U_08_SIN', 'Zonal wind  8hr. sin coeff.',
                                 0, 'GDZ', 'sph', ['time', 'lon', 'lat',
                                                   'height'], 'm/s'],
                  'U_12_COS': ['U_12_COS_ilev', 'Zonal wind 12hr. cos coeff.',
                               0, 'GDZ', 'sph', ['time', 'lon', 'lat', 'ilev'],
                               'm/s'],
                  'U_12_COS_2': ['U_12_COS', 'Zonal wind 12hr. cos coeff.',
                                 0, 'GDZ', 'sph', ['time', 'lon', 'lat',
                                                   'height'], 'm/s'],
                  'U_12_SIN': ['U_12_SIN_ilev', 'Zonal wind 12hr. sin coeff.',
                               0, 'GDZ', 'sph', ['time', 'lon', 'lat', 'ilev'],
                               'm/s'],
                  'U_12_SIN_2': ['U_12_SIN', 'Zonal wind 12hr. sin coeff.',
                                 0, 'GDZ', 'sph', ['time', 'lon', 'lat',
                                                   'height'], 'm/s'],
                  'U_24_COS': ['U_24_COS_ilev', 'Zonal wind 24hr. cos coeff.',
                               0, 'GDZ', 'sph', ['time', 'lon', 'lat', 'ilev'],
                               'm/s'],
                  'U_24_COS_2': ['U_24_COS', 'Zonal wind 24hr. cos coeff.',
                                 0, 'GDZ', 'sph', ['time', 'lon', 'lat',
                                                   'height'], 'm/s'],
                  'U_24_SIN': ['U_24_SIN_ilev', 'Zonal wind 24hr. sin coeff.',
                               0, 'GDZ', 'sph', ['time', 'lon', 'lat', 'ilev'],
                               'm/s'],
                  'U_24_SIN_2': ['U_24_SIN', 'Zonal wind 24hr. sin coeff.',
                                 0, 'GDZ', 'sph', ['time', 'lon', 'lat',
                                                   'height'], 'm/s'],
                  'V_08_COS': ['V_08_COS_ilev', 'Meridional wind  8hr. cos ' +
                               'coeff.', 0, 'GDZ', 'sph',
                               ['time', 'lon', 'lat', 'ilev'], 'm/s'],
                  'V_08_COS_2': ['V_08_COS', 'Meridional wind  8hr. cos ' +
                                 'coeff.', 0, 'GDZ', 'sph',
                                 ['time', 'lon', 'lat', 'height'], 'm/s'],
                  'V_08_SIN': ['V_08_SIN_ilev', 'Meridional wind  8hr. sin ' +
                               'coeff.', 0, 'GDZ', 'sph',
                               ['time', 'lon', 'lat', 'ilev'], 'm/s'],
                  'V_08_SIN_2': ['V_08_SIN', 'Meridional wind  8hr. sin ' +
                                 'coeff.', 0, 'GDZ', 'sph',
                                 ['time', 'lon', 'lat', 'height'], 'm/s'],
                  'V_12_COS': ['V_12_COS_ilev', 'Meridional wind 12hr. cos ' +
                               'coeff.', 0, 'GDZ', 'sph',
                               ['time', 'lon', 'lat', 'ilev'], 'm/s'],
                  'V_12_COS_2': ['V_12_COS', 'Meridional wind 12hr. cos ' +
                                 'coeff.', 0, 'GDZ', 'sph',
                                 ['time', 'lon', 'lat', 'height'], 'm/s'],
                  'V_12_SIN': ['V_12_SIN_ilev', 'Meridional wind 12hr. sin ' +
                               'coeff.', 0, 'GDZ', 'sph',
                               ['time', 'lon', 'lat', 'ilev'], 'm/s'],
                  'V_12_SIN_2': ['V_12_SIN', 'Meridional wind 12hr. sin ' +
                                 'coeff.', 0, 'GDZ', 'sph',
                                 ['time', 'lon', 'lat', 'height'], 'm/s'],
                  'V_24_COS': ['V_24_COS_ilev', 'Meridional wind 24hr. cos ' +
                               'coeff.', 0, 'GDZ', 'sph',
                               ['time', 'lon', 'lat', 'ilev'], 'm/s'],
                  'V_24_COS_2': ['V_24_COS', 'Meridional wind 24hr. cos ' +
                                 'coeff.', 0, 'GDZ', 'sph',
                                 ['time', 'lon', 'lat', 'height'], 'm/s'],
                  'V_24_SIN': ['V_24_SIN_ilev', 'Meridional wind 24hr. sin ' +
                               'coeff.', 0, 'GDZ', 'sph',
                               ['time', 'lon', 'lat', 'ilev'], 'm/s'],
                  'V_24_SIN_2': ['V_24_SIN', 'Meridional wind 24hr. sin ' +
                                 'coeff.', 0, 'GDZ', 'sph',
                                 ['time', 'lon', 'lat', 'height'], 'm/s']
                  }

# lists of known variables per file type
# assume h1 file is always present
gvar_keys = {'h0': ['RHO_CLUBB_lev'],  # also containes km_ilev
             'h1': ['co2vmr', 'ch4vmr', 'n2ovmr', 'f11vmr', 'f12vmr',
                    'sol_tsi', 'colat_crit1', 'colat_crit2', 'ED1', 'ED2',
                    'EDYN_ZIGM11_PED', 'EDYN_ZIGM2_HAL', 'ElecColDens', 'H',
                    'O', 'O2', 'OMEGA', 'PHIM2D', 'PS', 'T', 'TElec', 'TIon',
                    'U', 'UI', 'V', 'VI', 'WI', 'Z3', 'e', 'RHO_CLUBB',
                    'Z3GM'],
             # keys in h2 files that are not in the h1 files
             'h2': ['ED1', 'ED2', 'EDens', 'ElecColDens', 'HMF2', 'NMF2',
                    'OPLUS', 'TElec', 'TIon', 'UI', 'VI', 'WI', 'Z3'],
             # keys in h3 files that are not in the h1 files
             'h3': ['CO2', 'EDens', 'EKGW', 'N', 'NO', 'OpDens', 'QCO2',
                    'QHC2S', 'QJOULE', 'QNO', 'QO3', 'QO3P', 'QRS_TOT',
                    'SolIonRate_Tot', 'TTGW', 'UTGW_TOTAL', 'Z3GM'],
             # keys in the h4 files that are not in the h1 or h3 files
             'h4': ['OMEGA_08_COS', 'OMEGA_08_SIN', 'OMEGA_12_COS',
                    'OMEGA_12_SIN', 'OMEGA_24_COS', 'OMEGA_24_SIN',
                    'T_08_COS', 'T_08_SIN', 'T_12_COS', 'T_12_SIN',
                    'T_24_COS', 'T_24_SIN', 'U_08_COS', 'U_08_SIN',
                    'U_12_COS', 'U_12_SIN', 'U_24_COS', 'U_24_SIN',
                    'V_08_COS', 'V_08_SIN', 'V_12_COS', 'V_12_SIN',
                    'V_24_COS', 'V_24_SIN']}
# set up pressure level and height dependent equivalent variables
ilev_list = [value[0] for key, value in model_varnames.items()
             if value[5][-1] == 'ilev' if value not in
             ['H_geopot_ilev', 'H_geomet_ilev']]
ilev_replace = [item.split('_ilev')[0] for item in ilev_list if
                item not in ['H_geopot_ilev', 'H_geomet_ilev']]


# times from file converted to seconds since midnight of filedate
# plotting input times will be datetime strings of format 'YYYY-MM-DD HH:mm:ss'
# filedate is self.filedate from iri object
# converts to hours since midnight of filedate for plotting
def MODEL():

    from kamodo import Kamodo
    from glob import glob
    from os.path import isfile, getsize
    import psutil
    from netCDF4 import Dataset
    from numpy import array, NaN, zeros, where, unique, append, linspace
    from numpy import transpose, median
    from time import perf_counter
    from datetime import datetime, timezone
    import kamodo_ccmc.readers.reader_utilities as RU
    from waccmx_tocdf import convert_all

    class MODEL(Kamodo):
        '''WACCM-X model data reader.

        Inputs:
            file_dir: a string representing the file directory of the
                model output data.
                Note: This reader 'walks' the entire dataset in the directory.
            variables_requested = a list of variable name strings chosen from
                the model_varnames dictionary in this script, specifically the
                first item in the list associated with a given key.
                - If empty, the reader functionalizes all possible variables
                    (default)
                - If 'all', the reader returns the model_varnames dictionary
                    above for only the variables present in the given files.
            filetime = boolean (default = False)
                - If False, the script fully executes.
                - If True, the script only executes far enough to determine the
                    time values associated with the chosen data.
            printfiles = boolean (default = False)
                - If False, the filenames associated with the data retrieved
                    ARE NOT printed.
                - If True, the filenames associated with the data retrieved ARE
                    printed.
            gridded_int = boolean (default = True)
                - If True, the variables chosen are functionalized in both the
                    standard method and a gridded method.
                - If False, the variables chosen are functionalized in only the
                    standard method.
            verbose = boolean (False)
                - If False, script execution and the underlying Kamodo
                    execution is quiet except for specified messages.
                - If True, be prepared for a plethora of messages.
        All inputs are described in further detail in
            KamodoOnboardingInstructions.pdf.

        Returns: a kamodo object (see Kamodo core documentation) containing all
            requested variables in functionalized form.

        Notes:
            - WACCM-X model outputs are given in netCDF files, so no file
              conversion is needed. However, some of the variables depend on a
              pressure level coordinate, so some pre-processing is needed to
              calculate the median height across the entire dataset
              (time, lon, lat) for a given pressure level value for each type
              of pressure level coordinate (typically two).
            - WACCM-X data is given in several coordinate systems depending
              on pressure level - a representation of height in Pascals
              corresponding to a given atmospheric pressure. The preset values
              of pressure level in the coordinate systems are not
              guaranteed to be identical, so they are inverted independently of
              the other unless only one is provided. In that case, the are
              assumed to be identical.
            - Pressure level inversion is performed in the reader_utilities
              script, specifically the PLevelInterp function. Two versions of
              all variables that depend on pressure level are created: the
              original and one dependent on height, which is created through
              Kamodo's function composition feature.
            - Interpolation method 0 is chosen for the spatially independent
              variables, meaning the entire time series is read into memory and
              the standard SciPy interpolator is used.
            - All of the files have multiple time steps per file, but some of
              the files are small, while some are larger than 16 GB. This
              motivates some logic to automatically choose the appropriate
              interpolation method based on the file size compared to the
              available computer memory. Interpolation method 3 is chosen if
              the file size is large, and interpolation method 2 is chosen if
              the file size is manageable. In both cases, the standard SciPy
              interpolator is used.
        '''
        def __init__(self, file_dir, variables_requested=[],
                     printfiles=False, filetime=False, gridded_int=True,
                     verbose=False, **kwargs):
            super(MODEL, self).__init__(**kwargs)
            self.modelname = 'WACCMX'
            t0 = perf_counter()

            # first, check for file list, create if DNE
            list_file = file_dir + self.modelname + '_list.txt'
            time_file = file_dir + self.modelname + '_times.txt'
            self.times, self.pattern_files = {}, {}
            if not isfile(list_file) or not isfile(time_file):
                # collect filenames
                files = sorted(glob(file_dir+'*.h?.*.nc'))
                patterns = unique([file[-22:-20] for file in files])
                self.filename = ''.join([f+',' for f in files])[:-1]
                self.filedate = datetime.strptime(
                    files[0][-19:-9]+' 00:00:00', '%Y-%m-%d %H:%M:%S'
                    ).replace(tzinfo=timezone.utc)

                # check for h0 files, contains RHO_CLUBB_lev and km_ilev
                if 'h0' not in patterns:
                    self.conversion_test = convert_all(file_dir)
                    if not self.conversion_test:
                        return
                    patterns = append(patterns, 'h0')

                # establish time attributes
                for p in patterns:
                    # get list of files to loop through later
                    pattern_files = sorted(glob(file_dir+'*.'+p+'.*.nc'))
                    self.pattern_files[p] = pattern_files
                    self.times[p] = {'start': [], 'end': [], 'all': []}

                    # loop through to get times
                    for f in range(len(pattern_files)):
                        cdf_data = Dataset(pattern_files[f])
                        # hours since midnight first file
                        tmp = array(cdf_data.variables['time'])
                        pmt = array(cdf_data.variables['datesec'])
                        if f == 0:
                            int_time = int(tmp[0])  # date of first file
                        net = (tmp-int_time).astype(int)*24. + pmt/3600.  # hrs
                        self.times[p]['start'].append(net[0])
                        self.times[p]['end'].append(net[-1])
                        self.times[p]['all'].extend(net)
                        cdf_data.close()
                    # adjust all to be arrays
                    self.times[p]['start'] = array(self.times[p]['start'])
                    self.times[p]['end'] = array(self.times[p]['end'])
                    self.times[p]['all'] = array(self.times[p]['all'])

                # create time list file if DNE
                RU.create_timelist(list_file, time_file, self.modelname,
                                   self.times, self.pattern_files,
                                   self.filedate)
            else:  # read in data and time grids from file list
                self.times, self.pattern_files, self.filedate, self.filename =\
                    RU.read_timelist(time_file, list_file)
            if filetime:
                return  # return times only

            # perform initial check on variables_requested list
            if len(variables_requested) > 0 and variables_requested != 'all':
                test_list = [value[0] for key, value in model_varnames.items()]
                err_list = [item for item in variables_requested if item not in
                            test_list]
                if len(err_list) > 0:
                    print('Variable name(s) not recognized:', err_list)
                for item in err_list:
                    variables_requested.remove(item)
                if len(variables_requested) == 0:
                    return

            # collect variable list per file
            self.gvar_dict, gvar_list, self.err_list = {}, [], []
            # dict[key] = [gvar_list, file] for each type of file
            self.total_ilev = [item for item in ilev_list if item
                               not in ['H_geopot_ilev', 'H_geomet_ilev']]
            self.total_replace = ilev_replace
            # dictionary mapping to navigate related variable names
            self.ilev_map = {item1: item2 for item1, item2 in
                             zip(self.total_replace, self.total_ilev)}
            if len(variables_requested) > 0 and variables_requested != 'all':
                # add variables from each files if the file exists and if
                # variables are requested from that file.
                for key in gvar_keys.keys():
                    self.gvar_dict[key] = self.somefile_variables(
                        key, gvar_keys[key], variables_requested,
                        self.pattern_files)
                    gvar_list += self.gvar_dict[key][0]
                net_err = [value[0] for key, value in model_varnames.items()
                           if key not in gvar_list and value[0] in
                           self.err_list]
                if len(net_err) > 0:
                    print('Some requested variables are not available: ',
                          net_err)
            else:  # collect defaults
                # collect lists per file type
                for key in gvar_keys.keys():
                    self.gvar_dict[key] = self.allfile_variables(
                        key, gvar_keys[key], self.pattern_files)
                    gvar_list += self.gvar_dict[key][0]

                # returns list of variables included in data files
                if variables_requested == 'all':
                    self.var_dict = {value[0]: value[1:] for key, value in
                                     model_varnames.items() if key in
                                     gvar_list}
                    # add non-ilev versions of the variables in the files
                    key_list = list(self.var_dict.keys())
                    for var_key in key_list:
                        if var_key in self.total_ilev:
                            # retrieve equivalent non-ilev variable name
                            new_key = [key for key, value in
                                       self.ilev_map.items() if value ==
                                       var_key][0]
                            # retrieve key for model_varnames mapping
                            model_key = [key for key, value in
                                         model_varnames.items() if
                                         value[0] == new_key][0]
                            # add to returned dictionary
                            self.var_dict[model_varnames[model_key][0]] =\
                                model_varnames[model_key][1:]
                    return

            # store data for each variable desired and all coordinate grids
            # need to leave files open for variable data
            self.variables = {}
            for key in self.pattern_files.keys():
                # collect dimensions from first file
                cdf_datah = Dataset(self.pattern_files[key][0])
                for dim in cdf_datah.dimensions.keys():
                    # deal with dim renaming and skipping
                    if dim == 'ilev':
                        name = 'ilev1'
                    elif dim == 'lev':
                        name = 'ilev'
                    elif dim in ['lon', 'lat', 'mlon', 'mlat']:
                        name = dim
                    else:  # skip other extra string dimensions
                        continue  # time already calculated, don't overwrite

                    # prep for longitude wrapping
                    tmp = array(cdf_datah.variables[dim])
                    if (dim == 'lon'):
                        lon_le180 = list(where(tmp <= 180)[0])
                        lon_ge180 = list(where(tmp >= 180)[0])  # repeat 180
                        lon_idx = lon_ge180+lon_le180
                        setattr(self, '_lon_idx_'+key, lon_idx)
                        out = append(tmp, 360.) - 180.
                    elif dim == 'mlon':
                        lon_idx = append(linspace(0, len(tmp)-1, len(tmp),
                                                  dtype=int), 0)
                        out = append(tmp, 180.)
                        setattr(self, '_mlon_idx_'+key, lon_idx)
                    else:  # loop through other dimensions without changes
                        out = tmp
                    # store coordinate grids
                    setattr(self, '_'+name+'_'+key, out)
                cdf_datah.close()  # close for dimensions data since now arrays

                # loop through h0 files to calculate median km_ilev grid
                if 'h0' in key:
                    km_max, km_min, km = [], [], []
                    for f in self.pattern_files[key]:
                        cdf_datah0 = Dataset(f)
                        km.append(array(cdf_datah0.variables['km_ilev']))
                        km_max.append(cdf_datah0.km_ilev_max)
                        km_min.append(cdf_datah0.km_ilev_min)
                        cdf_datah0.close()
                    self._km_ilev = median(array(km), axis=0)
                    self._km_ilev_max = array(km_max).max()
                    self._km_ilev_min = array(km_min).min()

                # initialize variables dictionary
                for var in self.gvar_dict[key][0]:
                    self.variables[model_varnames[var][0]] = {
                        'units': model_varnames[var][-1], 'data': key}

            # store a few items
            self.missing_value = NaN
            if verbose:
                print(f'Took {perf_counter()-t0:.6f}s to read in data')
            if printfiles:
                print(self.filename)

            # register interpolators for each requested variable
            # rearrange to deal with H_ilev and H_ilev1 first if there
            varname_list = [key for key in self.variables.keys()]
            if 'H_geopot_ilev' in varname_list:
                varname_list.remove('H_geopot_ilev')
                varname_list = ['H_geopot_ilev'] + varname_list
            t_reg = perf_counter()
            for varname in varname_list:
                # register the variables
                self.register_variable(varname, gridded_int)
            if verbose:
                print(f'Took {perf_counter()-t_reg:.5f}s to register ' +
                      f'{len(varname_list)} variables.')
            if verbose:
                print(f'Took a total of {perf_counter()-t0:.5f}s to kamodofy' +
                      f' {len(varname_list)} variables.')

        # variable checks
        def variable_checks(self, variables_requested, g_varkeys_h,
                            cdf_datah_keys):
            '''Add pressure level variables if needed and perform variable
            checks for case where variables_requested is not empty.'''

            add_ilev = [var+'_ilev' for var in variables_requested if var
                        in self.total_replace]
            if len(add_ilev) > 0:
                add_ilev += ['H_geopot_ilev']
            new_var = variables_requested + add_ilev
            # remove replaced items and any items not expected to be in hx file
            short_var = [value[0] for key, value in model_varnames.items()
                         if value[0] in new_var and value[0] not in
                         self.total_replace and key in g_varkeys_h]
            gvar_listh = [key for key, value in model_varnames.items()
                          if value[0] in short_var and key
                          in cdf_datah_keys]

            # check for variables requested but not available
            if len(gvar_listh) != len(short_var):
                err_list = [value[0] for key, value in
                            model_varnames.items() if value[0] in
                            short_var and key not in cdf_datah_keys]
                for item in err_list:  # need a flat list
                    self.err_list.append(item)
                # add items to list, check full list later
            return gvar_listh

        # select desired variables from given file type
        def somefile_variables(self, h_type, g_varkeys_h, variables_requested,
                               file_dict):
            '''Add variables from desired file if file exists and if variables
            are requested.
            Inputs:
                - h_type is one of 'h2', 'h3', 'h4'.
                - g_varkeysh is a list of variables names known to exist in the
                    given file type.
                - variables_requested: a list of desired LaTeX variable names.
                - hx_files: a list of files for a given day.
            '''
            if h_type in file_dict.keys():
                cdf_datah = Dataset(file_dict[h_type][0])
                cdf_datah_keys = list(cdf_datah.variables.keys())
                cdf_datah.close()
                gvar_listh = self.variable_checks(variables_requested,
                                                  g_varkeys_h, cdf_datah_keys)
                return [gvar_listh, file_dict[h_type]]
            else:  # collect error list
                gvar_listh = self.variable_checks(variables_requested,
                                                  g_varkeys_h, [])
                return [gvar_listh, '']

        # return list of variables available in chosen file type
        def allfile_variables(self, h_type, g_varkeys_h, file_dict):
            '''Return list of variables from desired file if file exists.
            Inputs:
                - h_type is one of 'h2', 'h3', 'h4'.
                - g_varkeysh is a list of variables names known to exist in the
                    given file type.
            '''
            if h_type in file_dict.keys():
                cdf_datah = Dataset(file_dict[h_type][0])
                gvar_listh = [key for key, value in model_varnames.items()
                              if key in cdf_datah.variables.keys() and key in
                              g_varkeys_h]
                cdf_datah.close()
                return [gvar_listh, file_dict[h_type]]
            else:
                return [[], '']

        # dimension agnostic registration
        def register_variable(self, varname, gridded_int):
            '''Registers and functionalizes the dataset. Converts to a 1D
            numpy array for time series data.'''

            # create the coordinate dictionary
            key = self.variables[varname]['data']
            gvar = [key for key, value in model_varnames.items() if
                    value[0] == varname][0]  # variable name in file
            coord_list = [value[-2] for key, value in model_varnames.items()
                          if value[0] == varname][0]
            coord_str = [value[3]+value[4] for key, value in
                         model_varnames.items() if value[0] == varname][0]
            coord_dict = {'time': {'units': 'hr',
                                   'data': self.times[key]['all']}}
            if len(coord_list) == 1:  # convert 1D to array and functionalize
                data = []
                for f in self.pattern_files[key]:
                    cdf_data = Dataset(f)
                    tmp = array(cdf_data.variables[gvar])
                    cdf_data.close()
                    data.extend(tmp)
                self.variables[varname]['data'] = array(data)
                self = RU.Functionalize_Dataset(self, coord_dict, varname,
                                                self.variables[varname],
                                                gridded_int, coord_str,
                                                interp_flag=0)  # single array
                return
            for coord in coord_list[1:]:
                coord_dict[coord] = {'data': getattr(self,
                                                     '_'+coord+'_'+key)}
                if coord in ['lat', 'lon', 'mlon', 'mlat']:
                    coord_dict[coord]['units'] = 'deg'
                if coord in ['ilev', 'ilev1']:
                    coord_dict[coord]['units'] = 'hPa'
            lon_idx = getattr(self, '_'+coord_list[1]+'_idx_'+key)

            # overshooting guess at memory needed to read in all data
            # chunks/slices will be dropped if memory runs out later
            memory_needed = getsize(self.pattern_files['h1'][0]) * \
                len(self.pattern_files[key])
            memory_test = memory_needed < psutil.virtual_memory().available
            if varname == 'H_geopot_ilev':
                gridded_int = True
            if not memory_test:  # varname != 'H_geopot_ilev' and not
                # save memory by only retrieving time slices from the files
                # determine the operation to occur on each time slice
                def func(i, fi):  # i = file#, fi = slice#
                    cdf_data = Dataset(self.pattern_files[key][i])
                    tmp = array(cdf_data.variables[gvar][fi])
                    cdf_data.close()
                    # perform data wrangling on time slice
                    # (ilev,) lat/mlat, lon/mlon -> lon/mlon, lat/mlat, (ilev)
                    data = tmp.T[lon_idx]
                    return data

                # functionalize the 3D or 4D dataset, series of time slices
                self = RU.Functionalize_Dataset(
                    self, coord_dict, varname, self.variables[varname],
                    gridded_int, coord_str, interp_flag=3, func=func,
                    times_dict=self.times[key])
            else:  # only using this approach if there is enough memory
                # also using for H_ilev to speed up function composition
                # determine the operation to occur on each time chunk
                def func(f):  # file_idx
                    # convert slice number s to index in file
                    cdf_data = Dataset(self.pattern_files[key][f])
                    tmp = array(cdf_data.variables[gvar])
                    cdf_data.close()
                    # perform data wrangling on time chunk
                    # (ilev,) lat/mlat, lon/mlon -> lon/mlon, lat/mlat, (ilev)
                    if len(tmp.shape) == 4:
                        data = transpose(tmp, (0, 3, 2, 1))
                    elif len(tmp.shape) == 3:
                        data = transpose(tmp, (0, 2, 1))
                    return data[:, lon_idx]

                self = RU.Functionalize_Dataset(
                    self, coord_dict, varname, self.variables[varname],
                    gridded_int, coord_str, interp_flag=2, func=func,
                    times_dict=self.times[key])

            if len(coord_list) < 4:  # remaining logic is for 4D data.
                return

            # create pressure level -> km function once per ilev type
            if varname == 'H_geopot_ilev' or varname in self.total_ilev:
                if varname == 'H_geopot_ilev':
                    new_varname, units = 'Plev', 'hPa'
                    # perform unit conversion to km with Kamodo and invert
                    self['H_ilev_ijk[km]'] = 'H_geopot_ilev_ijk'
                    self[new_varname], interp_ijk = RU.PLevelInterp(
                        self['H_ilev_ijk'], coord_dict['time']['data'],
                        coord_dict['lon']['data'], coord_dict['lat']['data'],
                        coord_dict['ilev']['data'], units, self._km_ilev,
                        [self._km_ilev_max, self._km_ilev_min])
                    # max pressure level is at lowest height, so flip order
                elif varname != 'H_geopot_ilev':  # define by function comp
                    new_varname = varname.split('_ilev')[0]
                    units = self.variables[varname]['units']
                    self[new_varname] = varname+'(Plev)'
                    interp_ijk = self[new_varname]
                    self[new_varname].meta['arg_units'] = \
                        self['Plev'].meta['arg_units']
                self.variables[new_varname] = {'units': units, 'data': key}

                # create gridded interpolator if requested
                if gridded_int:
                    self = RU.register_griddedPlev(
                        self, new_varname, units, interp_ijk, coord_dict,
                        self._km_ilev, [self._km_ilev_max, self._km_ilev_min])
                    # max pressure level is at lowest height, so flip order
            return
    return MODEL



def figMods(fig, log10=False, lockAR=False, ncont=-1, colorscale='',
            cutInside=-1., returnGrid=False, enhanceHover=False, newTitle=''):
    '''
    Function to modify a plotly figure object in multiple ways.

    Arguments:
      fig         A plotly figure object
      log10       Logical, if true take the log of the contour values
      lockAR      Logical, if true lock the X/Y axis aspect ratio to 1
      ncont       number of contours to include in the plot
      colorscale  Set the desired colorscale for the plot
                  NOTE: appending '_r' to colorscale reverses it, ie RdBu_r
      cutInside   Make values inside a radius of cutInside NaNs
      returnGrid  Take the plot and return a new grid only plotly object
    '''

    import re
    import numpy as np
    import plotly.graph_objects as go

    if returnGrid:
        xx = fig.data[0]['x']
        yy = fig.data[0]['y']
        zz = fig.data[0]['z']
        xx_mg, yy_mg = np.meshgrid(xx, yy)
        xm = np.reshape(xx_mg, -1)
        ym = np.reshape(yy_mg, -1)
        fig2 = go.Figure(data=go.Scattergl(x=xm, y=ym, mode='markers'))
        return fig2

    if log10:
        # grab values from old plot
        val = fig.data[0]['z']
        # set negative and zero values to NaN
        # compute log10 of values, NaN will stay NaN
        val[val <= 0.] = np.nan
        val = np.log10(val)
        # assign back to plot object
        fig.data[0]['z'] = val
        # add log10 to colorbar label
        newlabel = 'log10<br>'+fig.data[0]['colorbar']['title']['text']
        fig.data[0]['colorbar']['title']['text'] = newlabel

    if lockAR:
        fig.update_xaxes(scaleanchor='y')

    if ncont > 0:
        fig.update_traces(ncontours=ncont,
                          contours=dict(coloring="fill", showlines=False))

    if colorscale != '':
        if colorscale == "BlueRed":
            fig.update_traces(colorscale="RdBu_r")
        elif colorscale == "Rainbow":
            fig.update_traces(
                colorscale=[[0.00, 'rgb(0,0,255)'],
                            [0.25, 'rgb(0,255,255)'],
                            [0.50, 'rgb(0,255,0)'],
                            [0.75, 'rgb(255,255,0)'],
                            [1.00, 'rgb(255,0,0)']])
        else:
            fig.update_traces(colorscale=colorscale)

    if cutInside > 0.:
        xx = fig.data[0]['x']
        yy = fig.data[0]['y']
        zz = fig.data[0]['z']
        xx_mg, yy_mg = np.meshgrid(xx, yy)  # Now the same dimensions as zz
        rr_mg = np.sqrt(xx_mg*xx_mg + yy_mg*yy_mg)
        maskR = rr_mg < cutInside
        zz[maskR] = np.nan
        fig.data[0]['z'] = zz

    if enhanceHover:
        xvar = re.sub(' +', ' ',
                      fig.layout.xaxis.title.text.strip('$').strip())
        yvar = re.sub(' +', ' ',
                      fig.layout.yaxis.title.text.strip('$').strip())
        cvar = re.sub(' +', ' ',
                      fig.data[0]['colorbar'].title.text.strip('$').strip())
        fig.update_traces(
            hovertemplate=xvar + ": %{x:.3f} <br>" +
            yvar + ": %{y:.3f} <br>" +
            cvar + ": %{z:.4g} <br>" +
            "<extra></extra>"
        )

    if newTitle != '':
        fig.layout.title.text = newTitle

    return fig


def toColor(fig, colorscale='Viridis'):
    '''
    Placeholder function for compatibility with old notebooks.
    Use figMods instead.

    Arguments:
      fig         A plotly figure object
      colorscale  Set the desired colorscale for the plot

    Returns a modified plotly figure object
    '''

    return figMods(fig, colorscale=colorscale, ncont=200)


def toLog10(fig):
    """
    Placeholder function for compatibility with old notebooks.
    Use figMods instead.

    Arguments:
      fig         A plotly figure object

    Returns a modified plotly figure object
    """

    return figMods(fig, log10=True)


def XYC(Xlabel, X, Ylabel, Y, Clabel, C, title='Plot Title',
        colorscale='Viridis', crange='',):
    """
    Simple 2D plot. Send in array of X points, Y points, and 2d array of
      values with labels for each and a basic plot is created.

    Arguments:
      Xlabel     Text label of X points
      X          1D array of points in X direction
      Ylabel     Text label of Y points
      Y          1D array of points in Y direction
      Clabel     Text label of plotted value
      C           2D array of values to plot, dimensioned[len(X),len(Y)]
      title       Optional plot title
      colorscale  Text string of colorscale to use for contours
      crange      Two position array with min/max contour values, [cmin,cmax]

    Returns a plotly figure object.
    """

    import numpy as np
    from kamodo import Kamodo

    # Compute/set contour range
    cmin = np.min(C)
    cmax = np.max(C)
    if crange != '':
        cmin = float(crange[0])
        cmax = float(crange[1])

    # Make basic figure to start from

    def plot_2D(X=X, Y=Y):
        return C

    plot1 = Kamodo(plot_2D=plot_2D)
    fig = plot1.plot(plot_2D=dict())

    # Set colorscale
    fig = figMods(fig, colorscale=colorscale, ncont=200)

    # Set plot options
    fig.update_traces(
        zmin=cmin, zmax=cmax, colorbar=dict(title=Clabel, tickformat=".3g"),
        hovertemplate=Xlabel + ": %{x:.4g}<br>" + Ylabel + ": %{y:.4g}<br>" +
        Clabel + ": %{z:.6g}<br>" + "<extra></extra>"
    )
    fig.update_layout(
        scene_aspectmode='data',
        title=dict(text=title,
                   yref="container", yanchor="top", x=0.01, y=0.97,
                   font=dict(size=16, family="sans serif", color="#000000")),
        margin=dict(l=0), xaxis_title=Xlabel, yaxis_title=Ylabel
    )

    return fig


def ReplotLL3D(figIn, model, altkm, plotts, plotCoord='GEO',
               title='Plot Title', colorscale='Viridis', crange='',
               opacity=0.70, axis=True, debug=0):
    """
    Takes a gridified 2D lon/lat figure and creates new plots
    in 3D and for chosen coordinate systems.
    """

    import numpy as np
    import datetime as dt
    import pytz
    from kamodo import Kamodo
    import kamodo_ccmc.flythrough.model_wrapper as MW
    from kamodo_ccmc.flythrough.utils import ConvertCoord
    from kamodo_ccmc.tools.shoreline import shoreline

    # Get current coordinate system and type. Create string for labels
    tmp = list(MW.Model_Variables(model, return_dict=True).values())[0][2:4]
    co = tmp[0]
    cot = tmp[1]

    # Pull out lon, lat, values and min/max from passed in figure
    lon = figIn.data[0]['x']
    lat = figIn.data[0]['y']
    val = figIn.data[0]['z']
    val2 = np.reshape(val, (len(lat), len(lon)))
    varn = figIn.data[0]['colorbar']['title']['text']
    cmin = np.min(val)
    cmax = np.max(val)
    if crange != '':
        cmin = float(crange[0])
        cmax = float(crange[1])

    # Prepare variables for use later
    rscale = (altkm + 6.3781E3)/6.3781E3
    lon_mg, lat_mg = np.meshgrid(np.array(lon), np.array(lat))
    x = lon_mg
    y = lat_mg
    if co == 'GDZ':
        z = np.full(x.shape, altkm)
    else:
        z = np.full(x.shape, rscale)
    t = np.full(x.shape, plotts)
    full_x = np.reshape(x, -1)
    full_y = np.reshape(y, -1)
    full_z = np.reshape(z, -1)
    full_t = np.reshape(t, -1)
    ilon = np.linspace(-180, 180, 181)
    # ilat = np.linspace(-90, 90, 91)

    # Convert incoming coordinates into plot coordinages (cartesian)
    if co == 'GDZ':
        # GDZ does not convert into other coordinates well, first go to GEO-car
        xx, yy, zz, units = ConvertCoord(full_t, full_x, full_y, full_z,
                                         co, cot, 'GEO', 'car')
        full_x, full_y, full_z = xx, yy, zz
        if plotCoord != 'GEO':
            xx, yy, zz, units = ConvertCoord(full_t, full_x, full_y, full_z,
                                             'GEO', 'car', plotCoord, 'car')
    else:
        xx, yy, zz, units = ConvertCoord(full_t, full_x, full_y, full_z,
                                         co, cot, plotCoord, 'car')
    x = np.reshape(xx, (len(lat), len(lon)))
    y = np.reshape(yy, (len(lat), len(lon)))
    z = np.reshape(zz, (len(lat), len(lon)))
    # NaNs in original plot don't hide properly, so set positions to NaN
    mask2 = np.isnan(val2)
    x[mask2] = np.nan
    y[mask2] = np.nan
    z[mask2] = np.nan

    # Generate initial figure to build upon

    def plot3d_var(X=x, Y=y, Z=z):
        return val2

    kobject = Kamodo(plot_var=plot3d_var)
    fig = kobject.plot(plot_var=dict())

    # Set colorscale
    if colorscale == "BlueRed":
        fig.update_traces(colorscale="RdBu_r")
    elif colorscale == "Rainbow":
        fig.update_traces(
            colorscale=[[0.00, 'rgb(0,0,255)'],
                        [0.25, 'rgb(0,255,255)'],
                        [0.50, 'rgb(0,255,0)'],
                        [0.75, 'rgb(255,255,0)'],
                        [1.00, 'rgb(255,0,0)']])
    else:
        fig.update_traces(colorscale=colorscale)

    # Set plot options
    fig.update_traces(
        cmin=cmin, cmax=cmax, colorbar=dict(title=varn, tickformat=".3g"),
        opacity=opacity,
        customdata=np.dstack((
            np.transpose(val2, axes=[1, 0]),
            np.transpose(lon_mg, axes=[1, 0]),
            np.transpose(lat_mg, axes=[1, 0]))),
        hovertemplate="<b>" + model + "</b><br>" +
        "X: %{x:.2f} " + units[0] + " " + plotCoord + "<br>"
        "Y: %{y:.2f} " + units[1] + " " + plotCoord + "<br>"
        "Z: %{z:.2f} " + units[2] + " " + plotCoord + "<br>"
        "Lon: %{customdata[1]:.2f} " + co + "<br>" +
        "Lat: %{customdata[2]:.2f} " + co + "<br>" +
        "Alt: " + "{:.2f}".format(altkm) + " km<br>" +
        varn + ": %{customdata[0]:.4g} <br>" +
        "<extra></extra>"
    )
    if not axis:
        fig.update_scenes(xaxis=dict(visible=False),
                          yaxis=dict(visible=False),
                          zaxis=dict(visible=False))
    plotdt = dt.datetime.fromtimestamp(plotts, tz=pytz.utc)
    timestr = plotdt.strftime("%Y/%m/%d %H:%M:%S")
    lltxt = model + ',  ' + plotCoord + ' Coordinates,  '
    lltxt = lltxt + str(altkm) + ' km Altitude,  ' + timestr
    fig.update_layout(
        scene_aspectmode='data',
        title=dict(text=title,
                   yref="container", yanchor="top", x=0.01, y=0.95,
                   font=dict(size=16, family="sans serif", color="#000000")),
        annotations=[
            dict(text=lltxt, x=0.0, y=0.0, ax=0, ay=0, xanchor="left",
                 xshift=0, yshift=-20, xref="paper", yref="paper",
                 font=dict(size=16, family="sans serif", color="#000000"))
        ],
        margin=dict(l=0),
        width=750,
        height=450,
    )

    # Add poles to plot
    xt = [0., 0.]
    yt = [0., 0.]
    zt = [-1.2, 1.2]
    fig.add_scatter3d(mode='lines', x=xt, y=yt, z=zt,
                      line=dict(width=4, color='black'),
                      showlegend=False,
                      hovertemplate=plotCoord+' Pole<extra></extra>')
    if plotCoord != 'GEO':
        # Add pole for GEO as well
        xt = np.array([0., 0.])
        yt = np.array([0., 0.])
        zt = np.array([-1., 1.])
        tt = np.full(xt.shape, plotts)
        xt, yt, zt, un = ConvertCoord(tt, xt, yt, zt,
                                      'GEO', 'car', plotCoord, 'car')
        xt = 1.2*xt
        yt = 1.2*yt
        zt = 1.2*zt
        fig.add_scatter3d(mode='lines', x=xt, y=yt, z=zt,
                          line=dict(width=4, color='rgb(79,79,79)'),
                          showlegend=False,
                          hovertemplate='GEO Pole<extra></extra>')

    # Zero latitude
    xt = rscale*(np.cos(ilon*np.pi/180.))
    yt = rscale*(np.sin(ilon*np.pi/180.))
    zt = np.zeros(xt.shape)
    fig.add_scatter3d(mode='lines', x=xt, y=yt, z=zt,
                      line=dict(width=2, color='black'),
                      showlegend=False,
                      hovertemplate=plotCoord+' Latitude=0<extra></extra>')

    # Create blank earth surface
    elon = np.linspace(-180, 180, 181)
    elat = np.linspace(-90, 90, 91)
    elon_mg, elat_mg = np.meshgrid(np.array(elon), np.array(elat))
    ex = -(np.cos(elat_mg*np.pi/180.)*np.cos(elon_mg*np.pi/180.))
    ey = -(np.cos(elat_mg*np.pi/180.)*np.sin(elon_mg*np.pi/180.))
    ez = (np.sin(elat_mg*np.pi/180.))
    colorse = np.zeros(shape=ex.shape)
    # colorse[ex<0.]=1  # Option to make day side a lighter color
    colorscalee = ['rgb(99,99,99)', 'rgb(0,0,0)']
    fig.add_surface(x=ex, y=ey, z=ez, surfacecolor=colorse,
                    cmin=0, cmax=1, colorscale=colorscalee,
                    showlegend=False, showscale=False, hoverinfo='skip')

    # Shoreline (land/water boundaries)
    pos = shoreline(rscale=1.001, coord=plotCoord, utcts=plotts)
    fig.add_scatter3d(mode='lines', x=pos[:, 0], y=pos[:, 1], z=pos[:, 2],
                      line=dict(width=2, color='white'),
                      showlegend=False, hoverinfo='skip')

    return fig


def GDZSlice4D(interp, varname, model, date, plotType, plotCoord='GEO',
               fixed_time='', fixed_lon='', fixed_lat='', fixed_alt='',
               title='Plot Title', shoreline=False, colorscale='Viridis',
               crange='', logscale=False):
    """
    Function takes 4D (Time/Lon/Lat/Alt) Kamodo interpolator in GDZ coordinates
    and creates a 2D plot by fixing two dimensions to create a slice in the
    other two dimensions.
    This can be displayed in any supported coordinate system.

    Arguments:
      interp      Kamodo interpolator for variable to plot
      varname     String name of variable to plot
      model       String name of model
      date        Datetime object for date of plot
      plotType    String with two variables to plot, ie. 'Lon-Lat'
                    choose from Time, Lon, Lat, Alt

    Optional Arguments:
      plotCoord   Coordinate system to display the plot in
      fixed_time  Fixed time value if not a plot axis [hours]
      fixed_lon   Fixed longitude value if not a plot axis [degrees]
      fixed_lat   Fixed latgitude value if not a plot axis [degrees]
      fixed_alt   Fixed altitude value if not a plot axis [km]
      title       String used for plot title
      shoreline   Logical to show shoreline on map if Lon-Lat plot
      colorscale  String name of colorscale to use
      crange      Contour range to override min,max, ie. [999,2001]
      logscale    Logical to set the colorscale range to log scale

    Returs a plotly figure object.
    """
    import numpy as np
    import time
    import datetime as dt
    from kamodo import Kamodo
    import kamodo_ccmc.flythrough.model_wrapper as MW
    from kamodo_ccmc.flythrough.utils import ConvertCoord
    from kamodo_ccmc.tools.shoreline import shoreline

    tic = time.perf_counter()

    # Check passed in plotType and parse for use later
    Types = ['Lon', 'Lat', 'Alt', 'Time']
    axis = plotType.split('-')
    if len(axis) != 2:
        print("ERROR, invalid plotType passed.")
        return
    if axis[0] not in Types or axis[1] not in Types:
        print("ERROR, plotType elements invalid.")
        return

    # Set base timestamp to start of day passed in.
    basets = date.replace(hour=0, minute=0, second=0,
                          microsecond=0).timestamp()
    if 'Time' not in axis:
        slicets = basets + fixed_time*3600.

    # Get current coordinate system and type from model. Create label strings
    tmp = list(MW.Model_Variables(model, return_dict=True).values())[0][2:4]
    co = tmp[0]
    cot = tmp[1]
    costr = co + '(' + cot + ')'
    if co != 'GDZ' or cot != 'sph':
        print('Expecting coordinates to be GDZ(sph) but got ', costr)

    # Set itim array
    if 'Time' in axis:
        itim = np.linspace(0, 24, 97)
        if axis[0] == 'Time':
            i1 = itim
        if axis[1] == 'Time':
            i2 = itim
        fixed_time = ''
    else:
        if fixed_time == '':
            print("ERROR, fixed_time not passed.")
            return
        itim = np.array([fixed_time])

    # Set ilon array
    if 'Lon' in axis:
        ilon = np.linspace(-180, 180, 91)
        if axis[0] == 'Lon':
            i1 = ilon
        if axis[1] == 'Lon':
            i2 = ilon
        fixed_lon = ''
    else:
        if fixed_lon == '':
            print("ERROR, fixed_lon not passed.")
            return
        ilon = np.array([fixed_lon])

    # Set ilat array
    if 'Lat' in axis:
        ilat = np.linspace(-90, 90, 73)
        if axis[0] == 'Lat':
            i1 = ilat
        if axis[1] == 'Lat':
            i2 = ilat
        fixed_lat = ''
    else:
        if fixed_lat == '':
            print("ERROR, fixed_lat not passed.")
            return
        ilat = np.array([fixed_lat])

    # Set ialt array
    if 'Alt' in axis:
        # Using quick bisection solve to find alt range.
        dh, h = 50., 50.
        v = interp([12, 0, 0, h])
        for _ in range(100):
            if np.isnan(v):
                if dh < 0.:
                    dh = -.5 * dh
            else:
                if abs(dh) < 1e-6:
                    break
                if dh > 0.:
                    dh = -.5 * dh
            h += dh
            v = interp([12, 0, 0, h])
        h1 = h
        dh = 200.
        for _ in range(100):
            if np.isnan(v):
                if dh > 0.:
                    dh = -.5 * dh
            else:
                if abs(dh) < 1e-6:
                    break
                if dh < 0.:
                    dh = -.5 * dh
            h += dh
            v = interp([12, 0, 0, h])
        ialt = np.linspace(h1, h, 101)
        if axis[0] == 'Alt':
            i1 = ialt
        if axis[1] == 'Alt':
            i2 = ialt
        fixed_alt = ''
    else:
        if fixed_alt == '':
            print("ERROR, fixed_alt not passed.")
            return
        ialt = np.array([fixed_alt])

    # Set grid to interpolate to
    i1mg, i2mg = np.meshgrid(np.array(i1), np.array(i2))
    w = np.reshape(i1mg, -1)  # Dummy to store shape for next steps

    if axis[0] == 'Lon':
        x = np.reshape(i1mg, -1)
    elif axis[1] == 'Lon':
        x = np.reshape(i2mg, -1)
    else:
        x = np.full(w.shape, fixed_lon)

    if axis[0] == 'Lat':
        y = np.reshape(i1mg, -1)
    elif axis[1] == 'Lat':
        y = np.reshape(i2mg, -1)
    else:
        y = np.full(w.shape, fixed_lat)

    z = np.ones(w.shape)  # Either 1km or 1Re are fine for coordinate transform

    # The t array needs to be timestamps
    if axis[0] == 'Time':
        t = np.reshape(i1mg, -1)
        t = basets + t*3600.
    elif axis[1] == 'Time':
        t = np.reshape(i2mg, -1)
        t = basets + t*3600.
    else:
        t = np.full(w.shape, slicets)

    # Convert from plotCoord coordinates into model coordinates to interpolate
    if plotCoord == 'GDZ':
        # No change to coordinates, so don't transform
        xx, yy, zz = x, y, z
    elif plotCoord == 'GEO':
        # Only altitude changes, and that is reset later
        xx, yy, zz = x, y, z
    else:
        # GDZ only converts to/from GEO, so  have to convert twice
        xx, yy, zz, units = ConvertCoord(t, x, y, z, plotCoord,
                                         'sph', 'GEO', 'sph')
        x, y, z = xx, yy, zz
        xx, yy, zz, units = ConvertCoord(t, x, y, z,
                                         'GEO', 'sph', 'GDZ', 'sph')

    # Don't use transformed zz, reset to desired altitude (km) to interpolate
    if axis[0] == 'Alt':
        zz = np.reshape(i1mg, -1)
    elif axis[1] == 'Alt':
        zz = np.reshape(i2mg, -1)
    else:
        zz = np.full(w.shape, fixed_alt)

    # The tt array needs to be in hours
    if axis[0] == 'Time':
        tt = np.reshape(i1mg, -1)
    elif axis[1] == 'Time':
        tt = np.reshape(i2mg, -1)
    else:
        tt = np.full(w.shape, fixed_time)

    positions = np.transpose([tt, xx, yy, zz])
    val = interp(positions)

    if logscale:
        val[val <= 0.] = np.nan
        val = np.log10(val)
        varname = 'log10<br>'+varname

    val2 = np.reshape(val, i1mg.shape)

    cmin = np.min(val)
    cmax = np.max(val)
    if crange != '':
        cmin = float(crange[0])
        cmax = float(crange[1])

    def plot_2D(v1=i1, v2=i2):
        return val2
    plot1 = Kamodo(plot_2D=plot_2D)
    fig = plot1.plot(plot_2D=dict())

    # Set colorscale
    fig = figMods(fig, colorscale=colorscale, ncont=200)

    # Set plot options
    fig.update_traces(
        zmin=cmin, zmax=cmax, colorbar=dict(title=varname, tickformat=".3g"),
        hovertemplate=str(axis[0]) + ": %{x:.2f}<br>" +
        str(axis[1]) + ": %{y:.2f}<br>" +
        varname + ": %{z:.6g}<br>" + "<extra></extra>"
    )
    fig.update_xaxes(title=axis[0])
    fig.update_yaxes(title=axis[1])
    if axis[0] == 'Lon' or axis[0] == 'Lat':
        fig.update_xaxes(tick0=0., dtick=45.)
    if axis[0] == 'Time':
        fig.update_xaxes(tick0=0., dtick=3.)
    if axis[1] == 'Lon' or axis[1] == 'Lat':
        fig.update_yaxes(tick0=0., dtick=45.)
    if axis[1] == 'Time':
        fig.update_yaxes(tick0=0., dtick=3.)
    timestr = date.strftime("%Y/%m/%d")
    # Check if fixed_time is >= 24 hours, if so adjust timestr and fixed_time
    if fixed_time != "":
        date2 = date
        while fixed_time >= 24.:
            fixed_time += -24.
            date2 = date2 + dt.timedelta(days=1)
            timestr = date2.strftime("%Y/%m/%d")
    subt = model + ',  ' + plotCoord + ' Coordinates,  '
    subt += timestr + ',  Slice at:'
    if fixed_time != "":
        subt = subt + '  ' + str(fixed_time) + ' hrs'
    if fixed_alt != "":
        subt = subt + '  ' + str(fixed_alt) + ' km Altitude'
    if fixed_lon != "":
        subt = subt + '  ' + str(fixed_lon) + ' deg Longitude'
    if fixed_lat != "":
        subt = subt + '  ' + str(fixed_lat) + ' deg Latitude'
    fig.update_layout(
        plot_bgcolor="white", scene_aspectmode='data',
        title=dict(text=title + '<br>' + subt,
                   yref="container", yanchor="top", x=0.01, y=0.97,
                   font=dict(size=16, family="sans serif", color="#000000")),
        margin=dict(l=0), width=750, height=375)

    if plotType == 'Lon-Lat':
        # Shoreline (land/water boundaries)
        if plotCoord == 'GDZ':
            # Convert into GEO as we only use lon/lat
            pos = shoreline(coord='GEO', coordT='sph', utcts=slicets)
        else:
            pos = shoreline(coord=plotCoord, coordT='sph', utcts=slicets)
        fig.add_scattergl(mode='lines', x=pos[:, 0], y=pos[:, 1],
                          hoverinfo='skip', showlegend=False,
                          line=dict(width=1, color='#EEEEEE'))

    toc = time.perf_counter()
    print(f"Time: {toc - tic:0.4f} seconds")
    return fig


def swmfgm3D(ko, var, time=0., title='',
             xcut=-999., ycut=-999., zcut=-999.,
             zgrid=False, znodes=False):
    '''
    Function to combine default X,Y,Z constant slices into one 3D
      plot. Returns a full 3D plotly figure.

    ko:      Kamodo object
    var:     string variable name
    time:    floating time in hours from start of first day of data
    title:   string text for plot title top left
    xcut:    if set, value for X slice through data
    ycut:    if set, value for Y slice through data
    zcut:    if set, value for Z slice through data
    zgrid:   logical to return plot of grid dz size
    znodes:  logical to show dots at grid locations
    '''
    import re
    import numpy as np
    import pandas as pd
    from kamodo import Kamodo
    import plotly.graph_objs as go

    useX = False
    useY = False
    useZ = False
    if xcut > -999.:
        useX = True
    if ycut > -999.:
        useY = True
    if zcut > -999.:
        useZ = True

    cmin = 1.e+99
    cmax = -1.e+99
    opacity = 1.

    if useZ:
        print('Extracting z slice at', zcut)
        figZ = ko.plot(var, plot_partial={var: {'time': time, 'Z': zcut}})
        x0 = figZ.data[0]['x']
        y0 = figZ.data[0]['y']
        z0 = np.array([zcut])
        val0 = figZ.data[0]['z']
        cmin = min(cmin, np.min(val0))
        cmax = max(cmax, np.max(val0))

    if useY:
        print('Extracting y slice at', ycut)
        figY = ko.plot(var, plot_partial={var: {'time': time, 'Y': ycut}})
        x1 = figY.data[0]['x']
        z1 = figY.data[0]['y']
        y1 = np.array([ycut])
        val1 = figY.data[0]['z']
        cmin = min(cmin, np.min(val1))
        cmax = max(cmax, np.max(val1))

    if useX:
        print('Extracting x slice at', xcut)
        figX = ko.plot(var, plot_partial={var: {'time': time, 'X': xcut}})
        y2 = figX.data[0]['x']
        z2 = figX.data[0]['y']
        x2 = np.array([xcut])
        val2 = figX.data[0]['z']
        cmin = min(cmin, np.min(val2))
        cmax = max(cmax, np.max(val2))

    print('starting plot')
    fig = go.Figure()
    if useZ:
        if zgrid:
            x_mg, y_mg, z_mg = np.meshgrid(x0, y0, z0)
            x_mg_1d = x_mg.reshape(-1)
            y_mg_1d = y_mg.reshape(-1)
            z_mg_1d = z_mg.reshape(-1)
            iv = []
            jv = []
            kv = []
            dx = []
            for iy in range(len(y0)-1):
                for ix in range(len(x0)-1):
                    # For each cell, create two triangles
                    iv.append(ix + iy*len(x0))
                    jv.append(ix+1 + iy*len(x0))
                    kv.append(ix + (iy+1)*len(x0))
                    dx.append(max(x_mg_1d[iv[-1]],
                                  x_mg_1d[jv[-1]],
                                  x_mg_1d[kv[-1]]) -
                              min(x_mg_1d[iv[-1]],
                                  x_mg_1d[jv[-1]],
                                  x_mg_1d[kv[-1]]))
                    iv.append(ix + (iy+1)*len(x0))
                    jv.append(ix+1 + (iy+1)*len(x0))
                    kv.append(ix+1 + iy*len(x0))
                    dx.append(max(x_mg_1d[iv[-1]],
                                  x_mg_1d[jv[-1]],
                                  x_mg_1d[kv[-1]]) -
                              min(x_mg_1d[iv[-1]],
                                  x_mg_1d[jv[-1]],
                                  x_mg_1d[kv[-1]]))
            dxlog2 = np.log2(dx)
            figG = go.Figure(data=[
                go.Mesh3d(
                    x=x_mg_1d,
                    y=y_mg_1d,
                    z=z_mg_1d,
                    colorbar_title='log2(dx)',
                    colorscale='Viridis',
                    intensity=dxlog2,
                    intensitymode='cell',
                    # intensity=x_mg_1d,  # If using, fix hover and labeling
                    # intensitymode='vertex',
                    i=iv,
                    j=jv,
                    k=kv,
                    name='y',
                    showscale=True
                )
            ])
            figG.update_traces(
                flatshading=True,
                customdata=dx,
                hovertemplate="<b>Grid Size</b><br>" +
                # "i: %{i:9i}<br>" +
                # "j: %{j:9i}<br>" +
                # "k: %{k:9i}<br>" +
                "log2(dx): %{intensity:.4g}<br>" +
                "dx: %{customdata:.4g}<br>" +
                "<extra></extra>"
            )
            if znodes:
                figG.add_scatter3d(x=x_mg_1d, y=y_mg_1d, z=z_mg_1d,
                                   mode='markers',
                                   marker=dict(size=1, color='black'),
                                   line=dict(width=1))

            return figG

        val0b = np.reshape(val0, (len(y0), len(x0), len(z0)))
        xvar = re.sub(' +', ' ',
                      figZ.layout.xaxis.title.text.strip('$').strip())
        yvar = re.sub(' +', ' ',
                      figZ.layout.yaxis.title.text.strip('$').strip())
        zvar = xvar.replace('X', 'Z')
        cvar = re.sub(' +', ' ',
                      figZ.data[0]['colorbar'].title.text.strip('$').strip())

        def plot3d_var0(X=x0, Y=y0, Z=z0):
            return val0b

        ko0 = Kamodo(plot_var=plot3d_var0)
        figZ2 = ko0.plot(plot_var=dict())
        figZ2.update_traces(colorscale='Viridis')
        figZ2.update_traces(
            cmin=cmin, cmax=cmax,
            colorbar=dict(title=cvar, tickformat=".3g"),
            opacity=opacity,
            customdata=np.transpose(val0b, axes=[1, 0, 2]),
            hovertemplate="<b>Position Values</b><br>" +
            xvar + ": %{x:.3f}<br>" +
            yvar + ": %{y:.3f}<br>" +
            zvar + ": %{z:.3f}<br>" +
            cvar + ": %{customdata:.4g}<br>" +
            "<extra></extra>"
        )
        fig.add_trace(figZ2.data[0])

    if useY:
        xm, ym, zm = np.meshgrid(x1, y1, z1)
        val1tp = np.transpose(val1, axes=[1, 0])
        val1b = np.reshape(val1tp, (len(y1), len(x1), len(z1)))
        xvar = re.sub(' +', ' ',
                      figY.layout.xaxis.title.text.strip('$').strip())
        zvar = re.sub(' +', ' ',
                      figY.layout.yaxis.title.text.strip('$').strip())
        yvar = xvar.replace('X', 'Y')
        cvar = re.sub(' +', ' ',
                      figY.data[0]['colorbar'].title.text.strip('$').strip())

        def plot3d_var1(X=x1, Y=y1, Z=z1):
            return val1b

        ko1 = Kamodo(plot_var=plot3d_var1)
        figY2 = ko1.plot(plot_var=dict())
        figY2.update_traces(colorscale='Viridis')
        figY2.update_traces(
            cmin=cmin, cmax=cmax,
            colorbar=dict(title=cvar, tickformat=".3g"),
            opacity=opacity,
            customdata=np.transpose(val1b, axes=[2, 1, 0]),
            hovertemplate="<b>Position Values</b><br>" +
            xvar + ": %{x:.3f} <br>" +
            yvar + ": %{y:.3f} <br>" +
            zvar+": %{z:.3f} <br>" +
            cvar+": %{customdata:.4g} <br>" +
            "<extra></extra>"
        )
        fig.add_trace(figY2.data[0])

    if useX:
        xm, ym, zm = np.meshgrid(x2, y2, z2)
        val2tp = np.transpose(val2, axes=[1, 0])
        val2b = np.reshape(val2tp, (len(y2), len(x2), len(z2)))
        yvar = re.sub(' +', ' ',
                      figX.layout.xaxis.title.text.strip('$').strip())
        zvar = re.sub(' +', ' ',
                      figX.layout.yaxis.title.text.strip('$').strip())
        xvar = yvar.replace('Y', 'Z')
        cvar = re.sub(' +', ' ',
                      figX.data[0]['colorbar'].title.text.strip('$').strip())

        def plot3d_var2(X=x2, Y=y2, Z=z2):
            return val2b

        ko2 = Kamodo(plot_var=plot3d_var2)
        figX2 = ko2.plot(plot_var=dict())
        figX2.update_traces(colorscale='Viridis')
        figX2.update_traces(
            cmin=cmin, cmax=cmax,
            colorbar=dict(title=cvar, tickformat=".3g"),
            opacity=opacity,
            customdata=np.transpose(val2b, axes=[2, 0, 1]),
            hovertemplate="<b>Position Values</b><br>" +
            xvar + ": %{x:.3f} <br>" + yvar + ": %{y:.3f} <br>" +
            zvar + ": %{z:.3f} <br>" + cvar + ": %{customdata:.4g} <br>" +
            "<extra></extra>"
        )
        fig.add_trace(figX2.data[0])

    # Create blank inner boundary
    elon = np.linspace(-180, 180, 181)
    elat = np.linspace(-90, 90, 91)
    elon_mg, elat_mg = np.meshgrid(np.array(elon), np.array(elat))
    ex = -2.5*(np.cos(elat_mg*np.pi/180.)*np.cos(elon_mg*np.pi/180.))
    ey = -2.5*(np.cos(elat_mg*np.pi/180.)*np.sin(elon_mg*np.pi/180.))
    ez = 2.5*(np.sin(elat_mg*np.pi/180.))
    colorse = np.zeros(shape=ex.shape)
    # colorse[ex<0.]=1  # Option to make day side a lighter color
    colorscalee = ['rgb(99,99,99)', 'rgb(0,0,0)']
    fig.add_surface(x=ex, y=ey, z=ez, surfacecolor=colorse,
                    cmin=0, cmax=1, colorscale=colorscalee,
                    showlegend=False, showscale=False, hoverinfo='skip')

    camera = dict(
        center=dict(x=0.35, y=0., z=0.),
    )
    fig.update_layout(
        scene_camera=camera,
        scene_aspectmode='data',
        title=dict(text=title,
                   yref="container", yanchor="top", x=0.01, y=0.97,
                   font=dict(size=16, family="sans serif", color="#000000"))
    )

    return fig


def swmfgm3Darb(ko, var, time=0., pos=[0, 0, 0], normal=[0, 1, 0],
                title='', lowerlabel='', showgrid=False, showibs=False,
                log10=False):
    '''
    Function to create a slice for any point and normal and interpolate
      from Kamodo object onto that grid. Returns a full 3D plotly figure.

    ko:    Kamodo object
    var:   string variable name
    time:  floating time in hours from start of first day of data
    pos, normal:   position and normal vector for slice plane
    title:         string text for plot title top left
    lowerlabel:    string text to label plot lower left
    showgrid:      logical to show dots at grid locations
    showibs:       logical to show inner boundary sphere
    log10:         logical to take log10 of display value
    '''
    import numpy as np
    import plotly.graph_objs as go
    import kamodo_ccmc.flythrough.model_wrapper as MW

    # Set interpolator and variable labels
    interp = getattr(ko, var)
    varu = ko.variables[var]['units']
    varlabel = var+" ["+varu+"]"

    # Compute values from pos, normal values
    uvec = normal/np.linalg.norm(normal)  # unit normal vector
    odist = np.dot(uvec, pos)  # closest distance to slice from origin
    opos = odist*uvec  # vector from origin to closest point

    # Compute base grid
    dg = np.linspace(-180, 180, 145)  # degree grid
    rv = 0.  # radius value
    if abs(odist) < 2.5:
        rv = 2.5*np.sin(((2.5-odist)/2.5)*np.pi/2.)
    rg = []  # radius grid
    rg.append(rv)
    rtrans = 2.5  # radius transition point from fixed dr to variable
    for _ in range(250):
        if rv < rtrans:
            rv += .11
        else:
            rv += .11*rv/rtrans
        if rv > 300.:
            break
        rg.append(rv)
    rg = np.array(rg)
    dg_mg, rg_mg = np.meshgrid(dg, rg)
    gx = rg_mg*np.cos(dg_mg*np.pi/180.)
    gy = rg_mg*np.sin(dg_mg*np.pi/180.)
    gx_1d = gx.reshape(-1)
    gy_1d = gy.reshape(-1)
    gz_1d = np.zeros([len(gx_1d)])
    time_1d = np.full((len(gx_1d)), time)
    grid0 = np.stack((gx_1d, gy_1d, gz_1d), axis=-1)  # nx3 position grid

    # Transform base grid to pos/normal and trim if needed
    # Rotate
    if abs(uvec[2]) < 1.:  # No rotation for +/- Z normal
        new_xaxis = np.cross([0, 0, 1], uvec)
        new_yaxis = np.cross(new_xaxis, uvec)
        transform = np.array([new_xaxis, new_yaxis, uvec]).T
        grid0 = np.inner(grid0, transform)
    # Shift
    grid0[:, 0] += opos[0]
    grid0[:, 1] += opos[1]
    grid0[:, 2] += opos[2]
    # back to 2D grid
    gx2 = grid0[:, 0].reshape(len(rg), -1)
    gy2 = grid0[:, 1].reshape(len(rg), -1)
    gz2 = grid0[:, 2].reshape(len(rg), -1)
    # Trim points beyond data, first extracting range
    cr = MW.Coord_Range(ko, [var], return_dict=True, print_output=False)
    x1, x2 = cr[var]['X'][0], cr[var]['X'][1]
    y1, y2 = cr[var]['Y'][0], cr[var]['Y'][1]
    z1, z2 = cr[var]['Z'][0], cr[var]['Z'][1]
    for j in range(len(dg)):
        for i in range(len(rg)):
            found = False
            frac = 0.
            if gx2[i, j] < x1:
                found = True
                frac = max(frac, (gx2[i, j] - x1)/(gx2[i, j] - gx2[i-1, j]))
            if gx2[i, j] > x2:
                found = True
                frac = max(frac, (gx2[i, j] - x2)/(gx2[i, j] - gx2[i-1, j]))
            if gy2[i, j] < y1:
                found = True
                frac = max(frac, (gy2[i, j] - y1)/(gy2[i, j] - gy2[i-1, j]))
            if gy2[i, j] > y2:
                found = True
                frac = max(frac, (gy2[i, j] - y2)/(gy2[i, j] - gy2[i-1, j]))
            if gz2[i, j] < z1:
                found = True
                frac = max(frac, (gz2[i, j] - z1)/(gz2[i, j] - gz2[i-1, j]))
            if gz2[i, j] > z2:
                found = True
                frac = max(frac, (gz2[i, j] - z2)/(gz2[i, j] - gz2[i-1, j]))
            if found:
                gx2[i, j] -= frac*(gx2[i, j] - gx2[i-1, j])
                gy2[i, j] -= frac*(gy2[i, j] - gy2[i-1, j])
                gz2[i, j] -= frac*(gz2[i, j] - gz2[i-1, j])
                gx2[i+1:, j] = gx2[i, j]
                gy2[i+1:, j] = gy2[i, j]
                gz2[i+1:, j] = gz2[i, j]
                break

    # Create 4D (nx4) grid and interpolate plot values
    grid = np.ndarray(shape=(len(gx_1d), 4), dtype=np.float32)
    grid[:, 0] = time_1d
    grid[:, 1] = gx2.reshape(-1)
    grid[:, 2] = gy2.reshape(-1)
    grid[:, 3] = gz2.reshape(-1)
    value = interp(grid)
    if log10:
        value[value <= 0.] = np.nan
        value = np.log10(value)
        varlabel = "log10("+varlabel+")"

    # Build connectivity grid cell by cell looping over positions
    iv = []
    jv = []
    kv = []
    for iy in range(len(rg)-1):
        for ix in range(len(dg)-1):
            # For each cell, create two triangular connectivity entries
            iv.append(ix+iy*len(dg))
            jv.append(ix+1+iy*len(dg))
            kv.append(ix+(iy+1)*len(dg))

            iv.append(ix+(iy+1)*len(dg))
            jv.append(ix+1+(iy+1)*len(dg))
            kv.append(ix+1+iy*len(dg))

    # Build resulting plot
    fig = go.Figure(data=[
        go.Mesh3d(
            x=grid[:, 1], y=grid[:, 2], z=grid[:, 3], i=iv, j=jv, k=kv,
            colorbar_title=varlabel, colorscale='Viridis',
            intensity=value, intensitymode='vertex',
            name='cont', showscale=True
        )
    ])
    fig.update_traces(
        flatshading=True,
        hovertemplate="X: %{x:.4g}<br>" + "Y: %{y:.4g}<br>" +
        "Z: %{z:.4g}<br>" + varlabel + ": %{intensity:.4g}<br>" +
        "<extra></extra>"
    )
    # Add grid points
    if showgrid:
        fig.add_scatter3d(
            x=grid[:, 1], y=grid[:, 2], z=grid[:, 3], mode='markers',
            marker=dict(size=1, color='white'), line=dict(width=1),
            name='grid'
        )

    # Create blank inner boundary
    if showibs:
        elon = np.linspace(-180, 180, 181)
        elat = np.linspace(-90, 90, 91)
        elon_mg, elat_mg = np.meshgrid(np.array(elon), np.array(elat))
        ex = -2.5*(np.cos(elat_mg*np.pi/180.)*np.cos(elon_mg*np.pi/180.))
        ey = -2.5*(np.cos(elat_mg*np.pi/180.)*np.sin(elon_mg*np.pi/180.))
        ez = +2.5*(np.sin(elat_mg*np.pi/180.))
        colorse = np.zeros(shape=ex.shape)
        # colorse[ex<0.]=1  # Option to make day side a lighter color
        colorscalee = ['rgb(99,99,99)', 'rgb(0,0,0)']
        fig.add_surface(x=ex, y=ey, z=ez, surfacecolor=colorse,
                        cmin=0, cmax=1, colorscale=colorscalee,
                        showlegend=False, showscale=False, hoverinfo='skip')

    fig.update_layout(
        scene_aspectmode='data',
        title=dict(text=title,
                   yref="container", yanchor="top", x=0.01, y=0.97,
                   font=dict(size=16, family="sans serif", color="#000000")),
        scene=dict(
            xaxis=dict(showbackground=False, showgrid=False),
            yaxis=dict(showbackground=False, showgrid=False),
            zaxis=dict(showbackground=False, showgrid=False)),
        annotations=[
            dict(text=lowerlabel, x=0.0, y=0.0, ax=0, ay=0, xanchor="left",
                 xshift=0, yshift=-20, xref="paper", yref="paper",
                 font=dict(size=16, family="sans serif", color="#000000"))
        ],
        margin=dict(l=0, t=35),
    )

    return fig
